from __future__ import annotations
from abc import ABC
from datetime import datetime, timedelta
from typing import Union, List, Any
import logging
from copy import deepcopy
import numpy as np
from collections import OrderedDict


class MeasurementBase(ABC):
    """
    Base class for all measurement types (Counts and Spectrum so far).

    Attributes
    ----------
    live_time : float
        The live time of the measurement
    real_time : float
        The real time of the measurement
    name : str
        Optional name to attach to the measurement.
    comment : str
        Optional comment to attach to the measurement.
    start_time : datetime
        Optional time at which the measurement began.
    log : logging.Logger
        Logger for logging loggable log logs.

    Methods
    -------
    end_time() -> datetime or None
        Property returning the end time of the measurement. "The end times are nigh."
    copy_me() -> Measurement
        Method which returns a deep copy of the measurement.
    resample(factor: float = 1.0) -> Measurement
        Must be implemented by subclasses.
    subtract_background(background: Measurement) -> Measurement
        Method which subtracts a background signal from the measurement.
    consecutive_addition(other: MeasurementBase) -> Measurement
        Must be implemented by subclasses.
    simultaneous_addition(other: MeasurementBase) -> Measurement
        Must be implemented by subclasses.
    consecutive_subtraction(other: MeasurementBase) -> Measurement
        Must be implemented by subclasses.
    simultaneous_subtraction(other: MeasurementBase) -> Measurement
        Must be implemented by subclasses.
    """
    def __init__(self,
                 live_time: float = 1.0,
                 real_time: float = 1.0,
                 name: str = '',
                 comment: str = '',
                 start_time: Union[datetime, None] = None):
        """
        Parameters
        ----------
        live_time : float
            The live time of the measurement
        real_time : float
            The real time of the measurement
        name : str
            Optional name to attach to the measurement.
        comment : str
            Optional comment to attach to the measurement.
        start_time : datetime
            Optional time at which the measurement began.
        """
        self.live_time = live_time
        self.real_time = real_time
        self.name = name
        self.comment = comment
        self.start_time = start_time
        self.log = logging.getLogger(f"{self.__class__.__name__}:{self.name}")
        self.log.debug(f"Created {self.__class__.__name__}({self.name})")

    @property
    def end_time(self) -> Union[datetime, None]:
        """
        Property which returns the time at which the measurement ended, if the start time is known, otherwise it
        returns None.

        Returns
        -------
        datetime
        """
        return None if self.start_time is None else self.start_time + timedelta(seconds=self.real_time)

    def copy_me(self):
        """
        Method which returns a deep copy of the measurement.

        Returns
        -------
        CopiedMeasurement
        """
        return deepcopy(self)

    def resample(self, factor: float = 1.0) -> MeasurementBase:
        raise NotImplementedError

    def subtract_background(self, background: MeasurementBase):
        """
        Method which subtracts a background signal from the measurement.

        Parameters
        ----------
        background : MeasurementBase

        Returns
        -------
        BackgroundSubtractedMeasurement
        """
        return self.simultaneous_subtraction(background)

    def consecutive_addition(self, other: MeasurementBase):
        raise NotImplementedError

    def simultaneous_addition(self, other: MeasurementBase):
        raise NotImplementedError

    def consecutive_subtraction(self, other: MeasurementBase):
        raise NotImplementedError

    def simultaneous_subtraction(self, other: MeasurementBase):
        raise NotImplementedError


class TickedMeasurement(MeasurementBase, ABC):
    """
    Base class for all ticked measurement types (TickedCounts and TickedSpectrum so far).

    Attributes
    ----------
    tick_times : Array[datetime]
        Numpy array of datetime objects, labelling the times at which the ticks occurred.
    live_times : Array[float]
        Numpy array of floats, determining the live times of each tick.
    real_times : Array[float]
        Numpy array of floats, determining the real times of each tick.
    live_time : float
        The total live time of the measurement
    real_time : float
        The total real time of the measurement
    name : str
        Optional name to attach to the measurement.
    comment : str
        Optional comment to attach to the measurement.
    start_time : datetime
        Optional time at which the measurement began.
    log : logging.Logger
        Logger for logging loggable log logs.

    Methods
    -------
    n_ticks() -> int
        Property which returns the number of ticks in the measurement.
    end_time() -> datetime or None
        Property returning the end time of the measurement. "The end times are nigh."
    copy_me() -> Measurement
        Method which returns a deep copy of the measurement.
    resample(factor: float = 1.0) -> Measurement
        Must be implemented by subclasses.
    subtract_background(background: Measurement) -> Measurement
        Method which subtracts a background signal from the measurement.
    consecutive_addition(other: MeasurementBase) -> Measurement
        Must be implemented by subclasses.
    simultaneous_addition(other: MeasurementBase) -> Measurement
        Must be implemented by subclasses.
    consecutive_subtraction(other: MeasurementBase) -> Measurement
        This method cannot be used for ticked types. Raises TypeError always.
    simultaneous_subtraction(other: MeasurementBase) -> Measurement
        Must be implemented by subclasses.
    get_single_tick(tick_number: int) -> Measurement
        Must be implemented by subclasses.
    sum_ticks(start: datetime = None, end: datetime = None) -> Measurement
        Must be implemented by subclasses.
    """
    def __init__(self,
                 tick_times: np.ndarray,
                 live_times: np.ndarray,
                 real_times: np.ndarray,
                 name: str = '',
                 comment: str = '',
                 start_time: Union[datetime, None] = None):
        """
        Parameters
        ----------
        tick_times : Array[datetime]
            Numpy array of datetime objects, labelling the times at which the ticks occurred.
        live_times : Array[float]
            Numpy array of floats, determining the live times of each tick.
        real_times : Array[float]
            Numpy array of floats, determining the real times of each tick.
        name : str
            Optional name to attach to the measurement.
        comment : str
            Optional comment to attach to the measurement.
        start_time : datetime
            Optional time at which the measurement began.
        """
        if start_time is None:
            start_time = tick_times[0]
        super().__init__(live_time=live_times.sum(),
                         real_time=real_times.sum(),
                         name=name,
                         comment=comment,
                         start_time=start_time)
        self.tick_times = tick_times
        self.live_times = live_times
        self.real_times = real_times

    @property
    def n_ticks(self) -> int:
        """
        Property which returns the number of ticks in the measurement.

        Returns
        -------
        int
        """
        return self.tick_times.size

    def get_single_tick(self, tick_number: int):
        raise NotImplementedError

    def sum_ticks(self, start: Union[datetime, None] = None, end: Union[datetime, None] = None):
        raise NotImplementedError

    def consecutive_subtraction(self, other: MeasurementBase):
        raise TypeError('Consecutive subtraction makes no sense for ticked types.')


class ArrayBase:
    """
    Base class for all Array types (CountsArray, SpectrumArray, TickedCountsArray and TickedSpectrumArray so far).

    Attributes
    ----------
    elements : OrderedDict[str, Measurement]
        OrderedDict, mapping measurements by their names.

    Methods
    -------
    names() -> List[str]
        Property which returns a list of the measurement names which are keys to the mapping.
    number_of_members() -> int
        Property which returns the number of members of the array.
    start_times() -> List[datetime]
        Property which returns a list of datetime instances which give the times at which each member of the array
        began to be collected.
    contains_all(names: List[str]) -> bool
        Method which returns True if all the strings in the argument match the names of elements in the array.
    names_match(other: ArrayBase) -> bool
        Method which returns True if all the names of the other array are present in this array, and vice-versa.
    consecutive_addition(other: ArrayBase) -> ArrayBase
        Method which returns the consecutive sum of the CountsArray with another Array.
    simultaneous_addition(other: ArrayBase) -> ArrayBase
        Method which returns the simultaneous sum of the CountsArray with another Array.
    consecutive_subtraction(other: ArrayBase) -> ArrayBase
        Method which returns the consecutive subtraction of another CountsArray from the Array.
    simultaneous_subtraction(other: ArrayBase) -> ArrayBase:
        Method which returns the simultaneous subtraction of another CountsArray from the Array.
    subtract_backgrounds(self, backgrounds: ArrayBase) -> ArrayBase:
        Method which subtracts the background from each member of the array, using the array of background counts
        supplied as the argument.
    simultaneous_sum() -> MeasurementBase
        Must be implemented by subclasses.
    consecutive_sum() -> MeasurementBase
        Must be implemented by subclasses.
    resample(factor: float) -> ArrayBase
        Method which returns a new array with each member resampled after scaling by the factor parameter.
    masked_array(names: List[str]) -> ArrayBase
        Method which returns a new array consisting only of members with the names in the given list of names.
    """
    def __init__(self, items):
        self.elements = OrderedDict([(i.name, i) for i in items])

    def __getitem__(self, item: str):
        return self.elements[item]

    def get(self, key: str, default_val: Any = None):
        return self.elements.get(key, default_val)

    def __setitem__(self, key: str, value):
        self.elements[key] = value

    def __contains__(self, item: str) -> bool:
        return item in self.elements

    def __iter__(self):
        for k in self.elements:
            yield self.elements[k]

    @property
    def names(self) -> List[str]:
        """
        Property which returns a list of the 'names' attributes for each member of the array.

        Returns
        -------
        List[str]
        """
        return list(self.elements.keys())

    @property
    def number_of_members(self) -> int:
        """
        Property which returns the number of members of the array.

        Returns
        -------
        int
        """
        return len(self.elements)

    @property
    def start_times(self) -> List[datetime]:
        """
        Property which returns a list of the 'start_time' attributes for each member of the array. The values are
        datetime objects or None.

        Returns
        -------
        List[Union[str, None]]
        """
        return [e.start_time for e in self]

    def contains_all(self, names: List[str]) -> bool:
        """
        Method which returns True if all the strings in the argument match the names of elements in the array.

        Parameters
        ----------
        names : List[str]
            List of names to match.

        Returns
        -------
        bool
        """
        return all([name in self for name in names])

    def names_match(self, other: ArrayBase):
        """
        Method which returns True if all the names of the other array are present in this array, and vice-versa.

        Parameters
        ----------
        other : Array

        Returns
        -------
        bool
        """
        return self.contains_all(other.names) and other.contains_all(self.names)

    def consecutive_addition(self, other: ArrayBase):
        """
        Method which returns the consecutive sum of the CountsArray with another Array.

        Parameters
        ----------
        other : Array
            The Array instance to add to this one.

        Returns
        -------
        Array
        """
        assert self.names_match(other)
        return self.__class__([d.consecutive_addition(other[d.name]) for d in self])

    def simultaneous_addition(self, other: ArrayBase):
        """
        Method which returns the simultaneous sum of the CountsArray with another Array.

        Parameters
        ----------
        other : Array
            The Array instance to add to this one.

        Returns
        -------
        Array
        """
        assert self.names_match(other)
        return self.__class__([d.simultaneous_addition(other[d.name]) for d in self])

    def consecutive_subtraction(self, other: ArrayBase):
        """
        Method which returns the consecutive subtraction of another CountsArray from the Array.

        Parameters
        ----------
        other : Array
            The Array instance to subtract from this one.

        Returns
        -------
        Array
        """
        assert self.names_match(other)
        return self.__class__([d.consecutive_subtraction(other[d.name]) for d in self])

    def simultaneous_subtraction(self, other: ArrayBase):
        """
        Method which returns the simultaneous subtraction of another CountsArray from the Array.

        Parameters
        ----------
        other : Array
            The Array instance to subtract from this one.

        Returns
        -------
        Array
        """
        assert self.names_match(other)
        return self.__class__([d.simultaneous_subtraction(other[d.name]) for d in self])

    def subtract_backgrounds(self, backgrounds: ArrayBase):
        """
        Method which subtracts the background from each member of the array, using the array of background counts
        supplied as the argument.

        Parameters
        ----------
        backgrounds : ArrayBase

        Returns
        -------
        Array
        """
        return self.simultaneous_subtraction(backgrounds)

    def simultaneous_sum(self):
        raise NotImplementedError

    def consecutive_sum(self):
        raise NotImplementedError

    def resample(self, factor: float):
        """
        Method which returns a new array with each member resampled after scaling by the factor parameter.

        Parameters
        ----------
        factor : float
            Scale by which the measurements will be scaled before resampling.

        Returns
        -------
        Array
        """
        return self.__class__([e.resample(factor) for e in self])

    def masked_array(self, names: List[str]):
        """
        Method which returns a new array consisting only of members with the names in the given list of names.

        Parameters
        ----------
        names : List[str]

        Returns
        -------
        Array
        """
        assert self.contains_all(names)
        return self.__class__([self[name] for name in names])


class TickedArrayBase(ArrayBase, ABC):
    """
    Base class for all TickedArray types (TickedCountsArray and TickedSpectrumArray so far).

    Attributes
    ----------
    elements : OrderedDict[str, Measurement]
        OrderedDict, mapping measurements by their names.

    Methods
    -------
    tick_times() -> Array[datetime]
        Property which returns a list of datetime instances labelling the time at which each tick occurred.
    n_ticks() -> int
        Property which returns the number of ticks in the first member of the array.
    n_ticks_all() -> List[int]
        Property which, like n_ticks, returns the number of ticks, but returns a list of values for each member, in
        case you suspect they may not all contain equal numbers.
    names() -> List[str]
        Property which returns a list of the measurement names which are keys to the mapping.
    number_of_members() -> int
        Property which returns the number of members of the array.
    start_times() -> List[datetime]
        Property which returns a list of datetime instances which give the times at which each member of the array
        began to be collected.
    contains_all(names: List[str]) -> bool
        Method which returns True if all the strings in the argument match the names of elements in the array.
    names_match(other: ArrayBase) -> bool
        Method which returns True if all the names of the other array are present in this array, and vice-versa.
    consecutive_addition(other: ArrayBase) -> ArrayBase
        Method which returns the consecutive sum of the CountsArray with another Array.
    simultaneous_addition(other: ArrayBase) -> ArrayBase
        Method which returns the simultaneous sum of the CountsArray with another Array.
    consecutive_subtraction(other: ArrayBase) -> ArrayBase
        This method makes no sense...
    simultaneous_subtraction(other: ArrayBase) -> ArrayBase:
        Method which returns the simultaneous subtraction of another CountsArray from the Array.
    subtract_backgrounds(self, backgrounds: ArrayBase) -> ArrayBase:
        Method which subtracts the background from each member of the array, using the array of background counts
        supplied as the argument.
    simultaneous_sum() -> MeasurementBase
        Must be implemented by subclasses.
    consecutive_sum() -> MeasurementBase
        Must be implemented by subclasses.
    resample(factor: float) -> ArrayBase
        Method which returns a new array with each member resampled after scaling by the factor parameter.
    masked_array(names: List[str]) -> ArrayBase
        Method which returns a new array consisting only of members with the names in the given list of names.
    get_single_tick(tick_num: int) -> ArrayBase
        Must be implemented by subclasses.
    sum_ticks(start: datetime, end: datetime) -> ArrayBase
        Must be implemented by subclasses.
    """
    def __init__(self, items: List[TickedMeasurement]):
        super(TickedArrayBase, self).__init__(items)

    @property
    def tick_times(self) -> np.ndarray:
        """
        Property which returns a list of datetime instances labelling the time at which each tick occurred (for the
        first member of the array).

        Returns
        -------
        Array[datetime]
        """
        return self[self.names[0]].tick_times

    @property
    def n_ticks(self) -> int:
        """
        Property which returns the number of ticks in the first member of the array.

        Returns
        -------
        int
        """
        return self.tick_times.size

    @property
    def n_ticks_all(self) -> List[int]:
        """
        Property which, like n_ticks, returns the number of ticks, but returns a list of values for each member, in
        case you suspect they may not all contain equal numbers.

        Returns
        -------
        List[int]
        """
        return [self[name].n_ticks for name in self.names]

    def get_single_tick(self, tick_num: int):
        raise NotImplementedError

    def sum_ticks(self, start: Union[datetime, None] = None, end: Union[datetime, None] = None):
        raise NotImplementedError



from __future__ import annotations

import os
import numpy as np
from datetime import datetime
from typing import Union, List, Tuple, Generator, Iterator, Any
from scipy.stats import norm
import pickle
import matplotlib.pyplot as plt
import struct
import requests

from geiger.core.calibration import Calibration
from geiger.core.base_classes import MeasurementBase, TickedMeasurement, ArrayBase, TickedArrayBase
from geiger.core.counts import Counts, CountsArray, TickedCounts, TickedCountsArray
from geiger.utilities.plotting import SpectrumPlotter


class Spectrum(MeasurementBase):
    """
    Class which stores a counts array and calibration for spectra.

    Parameters
    ----------
    energies : numpy.ndarray, Calibration or None
        Energy values for each channel, or energy calibration as a Calibration object. If None, an uncalibrated
        Calibration object is used (with a default of 1024 channels if no count data is provided).
    counts : numpy.ndarray or None
        Count values for each channel. If None, a numpy array of zeros is used, with the same number of channels as
        the energies or Calibration object used.
    live_time : float
        Live time in seconds. Default is 1.0 seconds.
    real_time : float
        Real time in seconds. Default is 1.0 seconds.
    name : str
        Label for the spectrum.
    bin_label : str
        'upper', 'middle' or 'lower'. If 'energies' argument is an energy array, this specifies what part of each
        bin each value labels, otherwise, this argument is ignored.
    comment : str
        Optional comment to attach to the spectrum.
    start_time : datetime or None
        Optional start time for the measurement.

    Attributes
    ----------
    calibration : Calibration
        Energy calibration as a Calibration object.
    counts : numpy.ndarray
        Count values for each channel.
    live_time : float
        Live time in seconds.
    real_time : float
        Real time in seconds.
    name : str
        Label for the spectrum.
    comment : str
        Optional comment to attach to the spectrum.
    start_time : datetime or None
        The time at which the acquisition began

    Methods
    -------
    energies -> np.ndarray
        Property which returns a numpy array of energy values which define the lower bin limits of the energy bins.
    n_channels -> int
        Property which returns the number of channels for the spectrum.
    channel_numbers -> np.ndarray
        Property which returns a numpy array of channel numbers.
    count_rate -> np.ndarray
        Property which returns a numpy array of count rate values for each channel.
    error -> np.ndarray
        Property which returns a numpy array of uncertainties for each channel.
    end_time -> datetime
        Property which returns the time at which the acquisition ended. Requires the start_time attribute to be filled.
    reset_poissonian_errors() -> Spectrum
        Method which rests the errors to the default Poisson values.
    find_channel(energy: float)-> int
        Method which returns the channel number into which a deposit of energy 'energy' would be counted.
    add_counts(energy: float, n_counts: float)
        Method which adds 'n_counts' counts of energy 'energy' to the spectrum.
    get_summed_counts(lower_lim, upper_lim, use_energy) -> Counts
        Method which returns a Counts instance, summed counts from the spectrum. 'lower_lim' and 'upper_lim' set the
        limits in channel numbers, or energies if 'use_energies' is set to True.
    rebin(new_calib: Calibration) -> Spectrum
        Method which rebins the spectrum to a new calibration.
    resample(scale: float = 1.0) -> Spectrum
        Method which uses the current spectrum (scaled by the 'scale' parameters) as the expectation for a new spectrum,
        sampled from either a normal or Poisson distribution.
    subtract_background(background_spectrum: Spectrum, negative: bool = True) -> Spectrum
        Method which subtracts a background spectrum (live time scaled) from the current spectrum, returning a new
        spectrum. The returned spectrum may contain negative values if the 'negative' parameter is True (default), but
        these are set to 0 if 'negative' is set to False.
    consecutive_addition(other: Spectrum) -> Spectrum
        Method which adds Count data to the current Spectrum, equivalent to adding two consecutive measurements
        together, adding their counts and live and real times together.
    simultaneous_addition(other: Spectrum) -> Spectrum
        Method which adds Count data to the current Spectrum, equivalent to adding two simultaneous measurements
        together, adding their counts together, scaling 'other' spectrum to have equal live time.
    consecutive_subtraction(other: Spectrum) -> Spectrum
        Method which subtracts Count data from the current Spectrum, equivalent to subtracting a subsection of a
        measurement. TBH this is never done, but I added the method for completeness.
    simultaneous_subtraction(other: Spectrum) -> Spectrum
        Method which subtracts Count data from the current Spectrum, equivalent to subtracting a signal present
        throughout the measurement, e.g. the background.
    copy_me() -> Spectrum
        Method which returns a deep  copy of the current spectrum.
    broaden(fwhms: np.ndarray) -> Spectrum
        Method which broadens the current spectrum, taking 'fwhms', an array of FWHM values for each channel of the
        spectrum.
    broaden_poly(fwhm_poly: Tuple[float, float, float, float]) -> Spectrum
        Method which broadens the current spectrum, using a polynomial (whose coefficients are given by the 'fwhm_poly'
        parameter). The equation is given by FWHM(E) = c0 + c1*E + c2*E**2 + c3*E**3.
    broaden_exp(fwhm_exp: Tuple[float, float, float]) -> Spectrum
        Method which broadens the current spectrum, using an exponential equation (whose coefficients are given by the
        'fwhm_exp' parameter). The equation is given by FWHM(E) = c0 + c1*E**c2.
    plot(*other_spectra: Spectrum, **kwargs)
        Method which plots the current spectrum and any other spectra passed as positional arguments. Keyword arguments
        are passed to the Multiplotter class.
    save(filename: str)
        Method which saves the current spectrum, using pickle, to the file name passed as an argument.
    load(filename: str) -> Spectrum
        Class method which loads a spectrum saved using the save() method above.
    from_txt(filename: str, **kwargs) -> Spectrum
        Class method which loads a spectrum in a txt file with energy and counts columns. Keyword arguments are passed
        to numpy.load_txt method which reads the file.
    """
    def __init__(self,
                 energies: Union[np.ndarray, Calibration, None] = None,
                 counts: Union[np.ndarray, None] = None,
                 live_time: float = 1.,
                 real_time: float = 1.,
                 error: Union[np.ndarray, None] = None,
                 name: str = '',
                 bin_label: str = 'lower',
                 comment: str = '',
                 start_time: Union[datetime, None] = None):
        """
        """
        if counts is None:
            if energies is None:
                self.calibration = Calibration.channel_numbers_only(1024)
            elif isinstance(energies, Calibration):
                self.calibration = energies
            elif isinstance(energies, np.ndarray):
                self.calibration = Calibration.from_energies(energies, bin_label)
            self.counts = np.zeros(self.calibration.n_channels)
        else:
            if energies is None:
                self.calibration = Calibration.channel_numbers_only(counts.size)
            elif isinstance(energies, Calibration):
                self.calibration = energies
            elif isinstance(energies, np.ndarray):
                self.calibration = Calibration.from_energies(energies, bin_label)
            self.counts = counts
        assert self.counts.size == self.calibration.n_channels

        if error is not None:
            assert error.size == self.n_channels

        self._error = error
        super(Spectrum, self).__init__(live_time=live_time, real_time=real_time, name=name, comment=comment,
                                       start_time=start_time)

    @property
    def energies(self) -> np.ndarray:
        """
        Property which returns a numpy array of energy values which define the lower bin limits of the energy bins.

        Returns
        -------
        np.ndarray
        """
        return self.calibration.energies

    @property
    def n_channels(self) -> int:
        """
        Property which returns the number of channels for the spectrum.

        Returns
        -------
        int
        """
        return self.calibration.n_channels

    @property
    def channel_numbers(self) -> np.ndarray:
        """
        Property which returns a numpy array of channel numbers.

        Returns
        -------
        np.ndarray
        """
        return self.calibration.channel_numbers

    @property
    def count_rate(self) -> np.ndarray:
        """
        Property which returns a numpy array of count rate values for each channel.

        Returns
        -------
        np.ndarray
        """
        return self.counts / self.live_time

    @property
    def error(self) -> np.ndarray:
        """
        Property which returns a numpy array of uncertainties for each channel.

        Returns
        -------
        np.ndarray
        """
        if self._error is None:
            assert (self.counts >= 0.).all()
            self._error = self.counts**0.5
            self.log.debug("Using Poisson errors")
        return self._error

    def reset_poissonian_errors(self) -> Spectrum:
        """
        Method which rests the errors to the default Poisson values.

        Errors inherited from convolution processes may be non-Poissonian.
        If we want to assume Poisson errors, call this method.

        Returns
        -------
        None
        """
        self._error = None
        return self

    def find_channel(self, energy: float) -> int:
        """
        Method which returns the channel number into which a deposit of energy 'energy' would be counted.

        Parameters
        ----------
        energy : float
            Energy of the deposit

        Returns
        -------
        int
        """
        return self.calibration.find_channel(energy)

    def add_counts(self, energy: float, n_counts: float = 1.0):
        """
        Method which adds 'n_counts' counts of energy 'energy' to the spectrum.

        Parameters
        ----------
        energy : float
            Energy of the deposit
        n_counts : int
            Number of deposits
        Returns
        -------
        None
        """
        channel = self.find_channel(energy)
        if channel is None:
            return
        self.counts[channel] += n_counts

    def get_summed_counts(self,
                          lower_lim: Union[int, float, None] = None,
                          upper_lim: Union[int, float, None] = None,
                          use_energy: bool = False
                          ) -> Counts:
        """
        Method which returns the total number of counts in the spectrum with optional upper and lower channel limits.
        The upper and lower limits can be specified as energy values if the 'use_energy' parameter is set to True.

        Parameters
        ----------
        lower_lim : int or float
            Lower channel (or lower energy) limit for count sum. Optional. Int by default. Float if using energy units.
        upper_lim : int or float
            Upper channel  (or lower energy) limit for count sum. Optional. Int by default. Float if using energy units.
        use_energy : bool
            Boolean switch determining whether energy units used in channel limiting.

        Returns
        -------
        Counts
        """

        c_mask = np.ones(self.n_channels, dtype=bool)
        if use_energy:
            if lower_lim is None:
                lower_lim = 0.
            if upper_lim is None:
                upper_lim = self.energies[-1] + self.calibration.coefficients[1]
            e0 = lower_lim
            de = upper_lim - lower_lim
            cal = Calibration(1, [e0, de, 0.])
            s = self.copy_me().rebin(cal)
            return s.get_summed_counts()
        if lower_lim is not None:
            c_mask = np.logical_and(c_mask, np.arange(self.n_channels) >= lower_lim)
        if upper_lim is not None:
            c_mask = np.logical_and(c_mask, np.arange(self.n_channels) < upper_lim)
        return Counts(self.counts[c_mask].sum(),
                      live_time=self.live_time,
                      real_time=self.real_time,
                      start_time=self.start_time,
                      error=np.sqrt(np.square(self.error[c_mask]).sum()),
                      name=self.name,
                      comment=self.comment)

    def rebin(self, new_calibration: Calibration) -> Spectrum:
        """
        Function which rebins spectrum with new calibration coefficients and number of channels.

        Parameters
        ----------
        new_calibration : Calibration
            Calibration of required rebinned spectrum.

        Returns
        -------
        new_spectrum : Spectrum
            Rebinned spectrum with calibration equal to 'new_calibration'.
        """

        # Get upper and lower energy values for each current channel
        energies_upper = self.calibration.upper_bin_limits
        energies_lower = self.calibration.lower_bin_limits

        # Get upper and lower energy values for each new channel
        new_energies_upper = new_calibration.upper_bin_limits
        new_energies_lower = new_calibration.lower_bin_limits

        # Generate empty array for new counts
        new_spectrum = self.copy_me()
        new_spectrum.calibration = new_calibration
        new_spectrum.counts = np.zeros(new_calibration.n_channels)
        new_spectrum._error = None if self._error is None else np.zeros(new_calibration.n_channels)

        # Check there will be counts in the new spectrum
        assert self.energies[0] < new_energies_upper[-1] and self.energies[-1] > new_energies_upper[0]

        zeros = np.zeros(new_calibration.n_channels)
        # Distribute the counts in each channel to the new spectrum
        for channel in range(self.n_channels):
            w_max = energies_upper[channel]
            w_min = energies_lower[channel]

            energy_overlap = np.maximum(np.minimum(new_energies_upper, w_max) -
                                        np.maximum(new_energies_lower, w_min),
                                        zeros)
            new_spectrum.counts += self.counts[channel] * energy_overlap / (w_max - w_min)
            if self._error is not None:
                new_spectrum._error += self._error[channel]**2 * energy_overlap / (w_max - w_min)

        if self._error is not None:
            new_spectrum._error = np.sqrt(new_spectrum._error)

        # Return the new spectrum
        return new_spectrum

    def resample(self, factor: float = 1.0) -> Spectrum:
        """
        Method which uses the current spectrum (scaled by the 'scale' parameters) as the expectation for a new spectrum,
        sampled from either a normal or Poisson distribution.

        Parameters
        ----------
        factor : float
            Multiplication factor for the expectations of the new spectral counts.

        Returns
        -------
        new_spectrum : Spectrum
            The resampled spectrum
        """
        new_spectrum = self.copy_me()
        if self._error is None:
            new_spectrum.counts[self.counts >= 0.] = np.random.poisson(self.counts[self.counts >= 0.]*factor)
        else:
            error = np.sqrt(np.square(self.error)*factor)
            new_spectrum.counts = np.random.normal(self.counts*factor, error)
            new_spectrum.counts[new_spectrum.counts < 0.] = 0.
            new_spectrum._error = error
        return new_spectrum

    def subtract_background(self, bg_spec: Spectrum, negative: bool = False) -> Spectrum:
        """
        Function which takes a background spectrum and subtracts it, scaled by the correct live time.

        Parameters
        ----------
        bg_spec : Spectrum class instance
            Background spectrum to be subtracted.
        negative : bool
            If set to False, the function sets all negative values in the subtracted spectrum to zero. Default: False.

        Returns
        -------
        Spectrum
        """
        sub = self.simultaneous_subtraction(bg_spec)
        if not negative:
            sub.counts[sub.counts < 0.] = 0.
        return sub

    def _check_other_spectrum(self, other: Spectrum) -> Spectrum:
        if other.calibration != self.calibration:
            other = other.rebin(self.calibration)
        return other

    def consecutive_addition(self, other: Spectrum) -> Spectrum:
        """
        Method which adds Count data to the current Spectrum, equivalent to adding two consecutive measurements
        together, adding their counts and live and real times together.

        Parameters
        ----------
        other : Spectrum
            Spectrum which should be added.

        Returns
        -------
        Spectrum
        """
        other = self._check_other_spectrum(other)
        return Spectrum(self.calibration, self.counts + other.counts,
                        live_time=self.live_time + other.live_time,
                        real_time=self.real_time + other.real_time,
                        start_time=self.start_time,
                        comment=self.comment,
                        error=np.sqrt(np.square(self.error) + np.square(other.error)),
                        name=self.name)

    def simultaneous_addition(self, other: Spectrum) -> Spectrum:
        """
        Method which adds Count data to the current Spectrum, equivalent to adding two simultaneous measurements
        together, adding their counts together, scaling 'other' spectrum to have equal live time.

        Parameters
        ----------
        other : Spectrum
            Spectrum which should be added

        Returns
        -------
        Spectrum
        """
        other = self._check_other_spectrum(other)
        scale = self.live_time / other.live_time
        return Spectrum(self.calibration,
                        self.counts + other.counts*scale,
                        live_time=self.live_time, real_time=self.real_time,
                        start_time=self.start_time,
                        comment=self.comment,
                        error=np.sqrt(np.square(self.error) + np.square(other.error*scale)),
                        name=self.name)

    def consecutive_subtraction(self, other: Spectrum) -> Spectrum:
        """
        Method which subtracts Count data from the current Spectrum, equivalent to subtracting a subsection of a
        measurement. TBH this is never done, but I added the method for completeness.

        Parameters
        ----------
        other : Spectrum
            Spectrum to be subtracted.

        Returns
        -------
        Spectrum
        """
        other = self._check_other_spectrum(other)
        return Spectrum(self.calibration, self.counts-other.counts,
                        live_time=self.live_time-other.live_time,
                        real_time=self.real_time-other.real_time,
                        start_time=self.start_time,
                        comment=self.comment,
                        error=np.sqrt(np.square(self.error) + np.square(other.error)),
                        name=self.name)

    def simultaneous_subtraction(self, other: Spectrum) -> Spectrum:
        """
        Method which subtracts Count data from the current Spectrum, equivalent to subtracting a signal present
        throughout the measurement, e.g. the background.

        Parameters
        ----------
        other : Spectrum
            Spectrum to be subtracted

        Returns
        -------
        Spectrum
        """
        other = self._check_other_spectrum(other)
        scale = self.live_time / other.live_time
        return Spectrum(self.calibration,
                        self.counts - other.counts*scale,
                        live_time=self.live_time, real_time=self.real_time,
                        start_time=self.start_time,
                        comment=self.comment,
                        error=np.sqrt(np.square(self.error) + np.square(other.error*scale)),
                        name=self.name)

    def __broaden(self, sigmas: np.ndarray) -> Spectrum:
        """
        Broaden spectrum with sigma (gaussian standard deviation) values in units of n_channels numbers for each
        channel.

        Parameters
        ----------
        sigmas : numpy.ndarray
            Array of sigma values.

        Returns
        -------
        Spectrum
        """

        def normal_dist(mu, sig, num):
            x = np.arange(num + 1)
            cdf = norm.cdf(x - 0.5, loc=mu, scale=sig)
            return cdf[1:] - cdf[:-1]

        assert isinstance(sigmas, np.ndarray)
        assert sigmas.size == self.n_channels

        new_spectrum = Spectrum(self.calibration, live_time=self.live_time, real_time=self.real_time, name=self.name)
        for ch, sigma in enumerate(sigmas):
            broad = normal_dist(ch, sigma, self.n_channels)
            new_spectrum.counts += self.counts[ch] * broad
        return new_spectrum

    def broaden(self, fwhms: np.ndarray) -> Spectrum:
        """
        Method which broadens the current spectrum, using an exponential equation (whose coefficients are given by the
        'fwhm_exp' parameter). The equation is given by FWHM(E) = c0 + c1*E**c2.

        Parameters
        ----------
        fwhms : numpy.ndarray
            Array of fwhm values.

        Returns
        -------
        Spectrum
        """
        assert isinstance(fwhms, np.ndarray)
        assert fwhms.size == self.n_channels

        delta_e = self.calibration.bin_widths
        sigmas = np.absolute(fwhms / (2.35 * delta_e))
        return self.__broaden(sigmas)

    def broaden_poly(self, fwhm_poly: Union[List, Tuple, np.ndarray]) -> Spectrum:
        """
        Method which broadens the current spectrum, using a polynomial (whose coefficients are given by the 'fwhm_poly'
        parameter). The equation is given by FWHM(E) = c0 + c1*E + c2*E**2 + c3*E**3.

        Parameters
        ----------
        fwhm_poly : numpy.ndarray, tuple or list
            Coefficients for broadening function.
        """
        assert len(fwhm_poly) == 4

        powers = np.arange(4)[np.newaxis, :]
        fwhms = (np.array(fwhm_poly)[np.newaxis, :] * np.power(self.energies[:, np.newaxis], powers)).sum(axis=1)
        new_spectrum = self.broaden(fwhms)
        new_spectrum.comment += "Broadened with polynomial coefficients: {}, {}, {}, {}; ".format(*fwhm_poly)
        return new_spectrum

    def broaden_exp(self, fwhm_exp: Union[List, Tuple, np.ndarray]) -> Spectrum:
        """
        Method which broadens the current spectrum, using an exponential equation (whose coefficients are given by the
        'fwhm_exp' parameter). The equation is given by FWHM(E) = c0 + c1*E**c2.

        Parameters
        ----------
        fwhm_exp : numpy.ndarray or list
            Coefficients for broadening function.

        Returns
        -------
        Spectrum
        """
        assert len(fwhm_exp) == 3

        fwhms = np.zeros(self.energies.size)
        fwhms[self.energies >= 0.] += fwhm_exp[1] * np.power(self.energies[self.energies >= 0.], fwhm_exp[2])
        fwhms += fwhm_exp[0]
        new_spectrum = self.broaden(fwhms)
        new_spectrum.comment += "Broadened with exponential coefficients: {}, {}, {}; ".format(*fwhm_exp)
        return new_spectrum

    def plot(self, *other_spectra: Spectrum, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """
        Method which plots the current spectrum and any other spectra passed as positional arguments. Keyword arguments
        are passed to the Multiplotter class.

        Parameters
        ----------
        other_spectra : Spectrum
            Other spectra to be plotted.
        kwargs : Keyword Arguments
            These will be passed to the Multiplotter class.

        Returns
        -------
        Pyplot figure and axes
        """
        return SpectrumPlotter(spectra=[self, *other_spectra], **kwargs).plot()

    def save(self, file_name: str):
        """
        Method which saves the current spectrum, using pickle, to the file name passed as an argument.

        Parameters
        ----------
        file_name : str
            The file name to which the spectrum will be saved.

        Returns
        -------
        None
        """
        calib_less_dict = dict((k, v) for k, v in self.__dict__.items() if k != 'calibration')
        with open(file_name, 'wb') as f:
            pickle.dump((self.calibration.__dict__, calib_less_dict), f)

    @classmethod
    def load(cls, file_name: str) -> Spectrum:
        """
        Class method which loads a spectrum saved using the save() method above.

        Parameters
        ----------
        file_name : str
            File name of the pickled spectrum.

        Returns
        -------
        Spectrum
        """
        with open(file_name, 'rb') as f:
            calib_dict, spec_dict = pickle.load(f)
        if 'n_channels' not in calib_dict:
            n_channels = calib_dict['channels']
        else:
            n_channels = calib_dict['n_channels']
        if 'coefficients' not in calib_dict:
            spec = cls(Calibration(n_channels, calib_dict['coeffs']))
        else:
            spec = cls(Calibration(calib_dict['n_channels'], calib_dict['coefficients']))
        spec.__dict__.update(spec_dict)
        return spec

    @classmethod
    def from_txt(cls, file_name: str, **kwargs) -> Spectrum:
        """
        Class method which generates a Spectrum object from counts and energy values in an ASCII text file.

        Parameters
        ----------
        file_name : str
            File name of text file.

        Other Parameters
        ----------------
        energy_column : int
            Column number of energy values. Optional, default = 0
        counts_column : int
            Column number of count values. Optional, default = 1
        skiprows : int
            Number of rows to be skipped at the top of the file. Optional,
            default = 0
        delim : str
            Delimiter of the columns in the ascii file. Optional, default is
            same as that of numpy.loadtxt function.
        """
        kwlib = dict(energy_column=0, counts_column=1, skiprows=0, calib=None, delim=None)
        kwlib.update(kwargs)
        e, c = np.loadtxt(file_name, delimiter=kwlib["delim"],
                          skiprows=kwlib["skiprows"],
                          usecols=(kwlib["energy_column"],
                                   kwlib["counts_column"])).T

        return cls(e, c, name=file_name.split("\\")[-1].split("/")[-1])


class EmissionSpectrum:
    @classmethod
    def from_lara_decay_data_file(cls,
                                  filename: str,
                                  calibration: Calibration = Calibration(1024, (0., 3., 0.))) -> Spectrum:
        """
        Class method which generates an emission spectrum for a specific
        isotope using data saved in the .lara format. E.g. available at
            http://www.nucleide.org/DDEP_WG/DDEPdata.htm

        Parameters
        ----------
        filename : str
            Filename of the .lara file to read
        calibration : Calibration
            Calibration instance specifying the calibration of the generated spectrum.
            Optional, default n_channels=1024, coeffs=(0., 3., 0.)

        Returns
        -------
            Spectrum object
        """
        with open(filename) as f:
            spectrum = cls._from_lara_decay_data(f.read().splitlines(), calibration)
        spectrum.comment = "Emission spectrum generated from: " + os.path.basename(filename)
        return spectrum

    @classmethod
    def from_lara_online(cls,
                         isotope: str = "Cs-137",
                         calibration: Calibration = Calibration(1024, [0., 3., 0.])) -> Spectrum:
        """
        Class method which generates an emission spectrum for a specific isotope using data collected from laraweb.

        Parameters
        ----------
        isotope : str
            Isotope in the format "Cs-137".
        calibration : Calibration
            Calibration for the generated spectrum.

        Returns
        -------
        Spectrum
        """
        data = requests.get(f"http://www.nucleide.org/Laraweb/Results/{isotope}.txt")
        return cls._from_lara_decay_data(data.text.splitlines(), calibration)

    @classmethod
    def _from_lara_decay_data(cls, data_lines: List[str], calibration: Calibration = Calibration(1024, [0., 3., 0.])):
        e = []
        i = []
        nuclide = (f"{data_lines.pop(0).split(';')[-1].strip()} "
                   f"({data_lines.pop(0).split(';')[-1].strip()}, "
                   f"Z = {data_lines.pop(0).split(';')[-1].strip()})")
        while " (keV) ; " not in data_lines.pop(0):
            pass
        while 1:
            line = data_lines.pop(0)
            if line.startswith("="):
                break
            line = line.replace(',', '.').split(" ; ")
            if line[4].startswith("g") or line[4].startswith("X"):
                e.append(float(line[0]))
                i.append(float(line[2]) / 100.)

        counts = np.zeros(calibration.n_channels)
        for j in range(len(e)):
            channel = calibration.find_channel(e[j])
            if channel is not None:
                counts[channel] += i[j]
        return Spectrum(calibration, counts, name=nuclide)


class MaestroSpectrum:
    @classmethod
    def chn_to_spectrum(cls, filename: str) -> Spectrum:
        """
        Class method which generates a Spectrum object from a Maestro .chn file.

        Parameters
        ----------
        filename : str
            Filename of the .chn file to read.
        """
        header_shape = '3h 2s 2i 8s 4s 2h'
        tail_shape = '2h 6f 228s B 63s B 63s 128s'

        with open(filename, 'rb') as f:
            header = struct.unpack(header_shape, f.read(32))
            spec = np.array(struct.unpack(f'{header[9]}i', f.read(4 * header[9])))
            info = struct.unpack(tail_shape, f.read(512))

        calibration = Calibration(header[9], (info[2], info[3], info[4]))
        return Spectrum(calibration, spec, live_time=header[5] * 0.02, real_time=header[4] * 0.02,
                        name=os.path.basename(filename))


class SpectrumArray(ArrayBase):
    """
    An array of Spectrum instances.

    Parameters
    ----------
    spectra : List[Spectrum]
        List of spectra to add to the array.

    Attributes
    ----------
    elements : OrderedDict[str, Spectrum]
        OrderedDict, mapping measurements by their names.

    Methods
    -------
    names() -> List[str]
        Property which returns a list of the measurement names which are keys to the mapping.
    number_of_members() -> int
        Property which returns the number of members of the array.
    start_times() -> List[datetime]
        Property which returns a list of datetime instances which give the times at which each member of the array
        began to be collected.
    contains_all(names: List[str]) -> bool
        Method which returns True if all the strings in the argument match the names of elements in the array.
    names_match(other: ArrayBase) -> bool
        Method which returns True if all the names of the other array are present in this array, and vice-versa.
    consecutive_addition(other: SpectrumArray) -> SpectrumArray
        Method which returns the consecutive sum of the CountsArray with another Array.
    simultaneous_addition(other: SpectrumArray) -> SpectrumArray
        Method which returns the simultaneous sum of the CountsArray with another Array.
    consecutive_subtraction(other: SpectrumArray) -> SpectrumArray
        Method which returns the consecutive subtraction of another CountsArray from the Array.
    simultaneous_subtraction(other: SpectrumArray) -> SpectrumArray
        Method which returns the simultaneous subtraction of another CountsArray from the Array.
    subtract_backgrounds(backgrounds: SpectrumArray) -> SpectrumArray:
        Method which subtracts the background from each member of the array, using the array of background counts
        supplied as the argument.
    simultaneous_sum() -> Spectrum
        Simultaneously sum all members of the array.
    consecutive_sum() -> Spectrum
        Consecutive sum of all members of the array.
    resample(factor: float) -> SpectrumArray
        Method which returns a new array with each member resampled after scaling by the factor parameter.
    masked_array(names: List[str]) -> SpectrumArray
        Method which returns a new array consisting only of members with the names in the given list of names.
    summed_counts_array(lower_lim: Union[int, float, None] = None,
                        upper_lim: Union[int, float, None] = None,
                        use_energy: bool = False) -> CountsArray
        Convert the SpectrumArray into a CountsArray, with optional upper and lower channel or energy bounds.
    plot(**kwargs):
        Plot the spectra on the same axes.
    """
    def __init__(self, spectra: List[Spectrum]):
        """
        """
        super(SpectrumArray, self).__init__(spectra)

    def __getitem__(self, item: str) -> Spectrum:
        return super(SpectrumArray, self).__getitem__(item)

    def get(self, item: str, default_val: Any = Spectrum()) -> Spectrum:
        return super().get(item, default_val)

    def __iter__(self) -> Iterator[Spectrum]:
        return super(SpectrumArray, self).__iter__()

    def consecutive_addition(self, other: SpectrumArray) -> SpectrumArray:
        """
        Consecutively add a SpectrumArray to this one.

        Parameters
        ----------
        other : SpectrumArray

        Returns
        -------
        SpectrumArray
        """
        return super(SpectrumArray, self).consecutive_addition(other)

    def consecutive_subtraction(self, other: SpectrumArray) -> SpectrumArray:
        """
        Consecutively subtract a SpectrumArray from this one.

        Parameters
        ----------
        other : SpectrumArray

        Returns
        -------
        SpectrumArray
        """
        return super(SpectrumArray, self).consecutive_subtraction(other)

    def simultaneous_addition(self, other: SpectrumArray) -> SpectrumArray:
        """
        Simultaneously add a SpectrumArray to this one.

        Parameters
        ----------
        other : SpectrumArray

        Returns
        -------
        SpectrumArray
        """
        return super(SpectrumArray, self).simultaneous_addition(other)

    def simultaneous_subtraction(self, other: SpectrumArray) -> SpectrumArray:
        """
        Simultaneously subtract a SpectrumArray from this one.

        Parameters
        ----------
        other: SpectrumArray

        Returns
        -------
        SpectrumArray
        """
        return super(SpectrumArray, self).simultaneous_subtraction(other)

    def simultaneous_sum(self, calibration: Calibration = None) -> Spectrum:
        """
        Simultaneously sum all members of the array.

        Parameters
        ----------
        calibration : Optional[Calibration]
            Calibration to use for the summed spectrum. If not specified, the first member of the array's calibration
            will be used.

        Returns
        -------
        Spectrum
        """
        if self.number_of_members == 0:
            return Spectrum(energies=calibration)
        if calibration is None:
            calibration = list(self.elements.values())[0].calibration
        spec = 0
        for s in self:
            if spec == 0:
                spec = s.rebin(calibration)
            else:
                spec = spec.simultaneous_addition(s)
        spec.name = ', '.join(self.names)
        return spec

    def consecutive_sum(self, calibration: Calibration = None) -> Spectrum:
        """
        Consecutive sum of all members of the array.

        Parameters
        ----------
        calibration : Optional[Calibration]
            Calibration to use for the summed spectrum. If not specified, the first member of the array's calibration
            will be used.

        Returns
        -------
        Spectrum
        """
        if self.number_of_members == 0:
            return Spectrum(energies=calibration)
        if calibration is None:
            calibration = list(self.elements.values())[0].calibration
        spec = 0
        for s in self:
            if spec == 0:
                spec = s.rebin(calibration)
            else:
                spec = spec.consecutive_addition(s)
        spec.name = ', '.join(self.names)
        return spec

    def summed_counts_array(self,
                            lower_lim: Union[int, float, None] = None,
                            upper_lim: Union[int, float, None] = None,
                            use_energy: bool = False
                            ) -> CountsArray:
        """
        Convert the SpectrumArray into a CountsArray, with optional upper and lower channel or energy bounds.

        Parameters
        ----------
        lower_lim : int or float
            Optional channel (default) or energy for lower bound of the sum.
        upper_lim : int or float
            Optional channel (default) or energy for upper bound of the sum.
        use_energy : bool
            Use energy values for the bounds, rather than channel number.

        Returns
        -------
        CountsArray
        """
        counts = [spec.get_summed_counts(lower_lim=lower_lim, upper_lim=upper_lim, use_energy=use_energy)
                  for spec in self]
        return CountsArray(counts)

    def plot(self, **kwargs):
        """
        Plot the spectra on the same axes.

        Parameters
        ----------
        kwargs
            Keyword arguments to pass to Multiplotter class.

        Returns
        -------
        plt.Figure, plt.Axes
        """
        return SpectrumPlotter(spectra=list(self.elements.values()), **kwargs).plot()

    def resample(self, factor: float) -> SpectrumArray:
        """
        Method which returns a new array with each member resampled after scaling by the factor parameter.

        Parameters
        ----------
        factor : float
            Scale by which the measurements will be scaled before resampling.

        Returns
        -------
        Array
        """
        return super(SpectrumArray, self).resample(factor)


class WindowSpectrumArray(SpectrumArray):
    """
    An array of Spectrum instances, generated by a rolling window.

    Parameters
    ----------
    spectra : List[Spectrum]
        List of spectra to add to the array.
    time : datetime
        Time label associated to the window.
    width : int
        Length of the window in ticks.

    Attributes
    ----------
    elements : OrderedDict[str, Spectrum]
        OrderedDict, mapping measurements by their names.
    time : datetime
        Time label associated to the window.
    window_width : int
        Length of the window in ticks.

    Methods
    -------
    names() -> List[str]
        Property which returns a list of the measurement names which are keys to the mapping.
    number_of_members() -> int
        Property which returns the number of members of the array.
    start_times() -> List[datetime]
        Property which returns a list of datetime instances which give the times at which each member of the array
        began to be collected.
    contains_all(names: List[str]) -> bool
        Method which returns True if all the strings in the argument match the names of elements in the array.
    names_match(other: ArrayBase) -> bool
        Method which returns True if all the names of the other array are present in this array, and vice-versa.
    consecutive_addition(other: SpectrumArray) -> SpectrumArray
        Method which returns the consecutive sum of the CountsArray with another Array.
    simultaneous_addition(other: SpectrumArray) -> SpectrumArray
        Method which returns the simultaneous sum of the CountsArray with another Array.
    consecutive_subtraction(other: SpectrumArray) -> SpectrumArray
        Method which returns the consecutive subtraction of another CountsArray from the Array.
    simultaneous_subtraction(other: SpectrumArray) -> SpectrumArray
        Method which returns the simultaneous subtraction of another CountsArray from the Array.
    subtract_backgrounds(backgrounds: SpectrumArray) -> SpectrumArray:
        Method which subtracts the background from each member of the array, using the array of background counts
        supplied as the argument.
    simultaneous_sum() -> Spectrum
        Simultaneously sum all members of the array.
    consecutive_sum() -> Spectrum
        Consecutive sum of all members of the array.
    resample(factor: float) -> SpectrumArray
        Method which returns a new array with each member resampled after scaling by the factor parameter.
    masked_array(names: List[str]) -> SpectrumArray
        Method which returns a new array consisting only of members with the names in the given list of names.
    summed_counts_array(lower_lim: Union[int, float, None] = None,
                        upper_lim: Union[int, float, None] = None,
                        use_energy: bool = False) -> CountsArray
        Convert the SpectrumArray into a CountsArray, with optional upper and lower channel or energy bounds.
    plot(**kwargs):
        Plot the spectra on the same axes.
    """
    def __init__(self, spectra: List[Spectrum], time: datetime, width: int = -1):
        super().__init__(spectra)

        # 'Time' label for the window
        self.time = time
        self.window_width = width


class TickedSpectrum(TickedMeasurement):
    """
    Spectrum class with ticks.

    Parameters
    ----------
    energies : numpy.ndarray, Calibration or None
        Energy values for each channel, or energy calibration as a Calibration object. If None, an uncalibrated
        Calibration object is used (with a default of 1024 channels if no count data is provided).
    tick_times : Array[datetime]
        Numpy array of datetime objects, labelling the times at which the ticks occurred.
    ticks : 2DArray[float]
        Counts in each energy bin for each tick.



    Attributes
    ----------
    ticks : 2DArray[float]
        Counts in each energy bin for each tick.
    calibration : Calibration
        Energy calibration as a Calibration object.
    tick_times : Array[datetime]
        Numpy array of datetime objects, labelling the times at which the ticks occurred.
    live_times : Array[float]
        Numpy array of floats, determining the live times of each tick.
    real_times : Array[float]
        Numpy array of floats, determining the real times of each tick.
    msg_times : np.ndarray
        Possible array of datetime instances, marking the times at which the tick messages arrived.
    blank_rate : np.ndarray
        Possible array of floats, marking the rate of blanking pulses measured during each tick.
    live_time : float
        The total live time of the measurement
    real_time : float
        The total real time of the measurement
    name : str
        Optional name to attach to the measurement.
    comment : str
        Optional comment to attach to the measurement.
    start_time : datetime
        Optional time at which the measurement began.
    log : logging.Logger
        Logger for logging loggable log logs.

    Methods
    -------
    energies() -> np.ndarray
        Property which returns the lower energy bound foreach channel in the spectrum.
    n_channels() -> int
        Property which returns the number of channels in each spectrum per tick.
    errors() -> np.ndarray
        Property which returns the uncertainties for each channel of each tick.
    n_ticks() -> int
        Property which returns the number of ticks in the measurement.
    end_time() -> datetime or None
        Property returning the end time of the measurement. "The end times are nigh."
    copy_me() -> TickedSpectrum
        Method which returns a deep copy of the measurement.
    resample(factor: float = 1.0) -> TickedSpectrum
        Must be implemented by subclasses.
    subtract_background(background: Spectrum) -> TickedSpectrum
        Method which subtracts a background signal from the measurement.
    consecutive_addition(other: TickedSpectrum) -> TickedSpectrum
        Method which consecutively sums another TickedSpectrum instance to this one.
    simultaneous_addition(other: Spectrum) -> TickedSpectrum
        Method which simultaneously sums a Spectrum or another TickedSpectrum to this TickedSpectrum.
    consecutive_subtraction(other: MeasurementBase) -> Measurement
        This method cannot be used for ticked types. Raises TypeError always.
    simultaneous_subtraction(other: Spectrum) -> TickedSpectrum
        Method which simultaneously subtracts a Spectrum or another TickedSpectrum from this TickedSpectrum.
    get_single_tick(tick_number: int) -> Spectrum
        Method which retrieves a single tick.
    sum_ticks(start: datetime = None, end: datetime = None) -> Spectrum
        Method which returns a summed spectrum with optional start and end
        times.
    get_ticked_counts(lower_lim: float = None, upper_lim: float = None, start: datetime = None, end: datetime = None,
                      use_energy: bool = False) -> TickedCounts
        Method which sums channels in each tick's spectrum to make a TickedCounts instance.
    simple_windowed_spectra_iterations(tick_window_width: int = 1, mode: int = 0) -> int
        Method which calculates the number of iterations for a rolling window of a given length and mode.
    simple_windowed_spectra(tick_window_width: int = 1, mode: int = 0) -> Generator[Spectrum]
        Simple generator of variable length windowed spectra, equivalent to a rolling window across the time space
    resample(factor: float = 1.0) -> TickedSpectrum
        Method which uses the current spectrum (scaled by the 'scale' parameters) as the expectation for a new spectrum,
        sampled from either a normal or Poisson distribution.
    rebin(new_calibration: Calibration) -> TickedSpectrum
        Function which rebins spectrum with new calibration coefficients and number of channels.
    """
    def __init__(self,
                 energies: Union[np.ndarray, Calibration],
                 tick_times: np.ndarray,
                 ticks: np.ndarray,
                 live_times: Union[np.ndarray, None] = None,
                 real_times: Union[np.ndarray, None] = None,
                 errors: Union[np.ndarray, None] = None,
                 name: str = '',
                 bin_label: str = 'lower',
                 comment: str = '',
                 start_time: Union[datetime, None] = None,
                 msg_times: Union[np.ndarray, None] = None,
                 blank_rate: Union[np.ndarray, None] = None):

        if live_times is None:
            n_gaps = (tick_times.size - 1)
            live_times = np.ones(tick_times.size) * (tick_times[-1] - tick_times[0]).total_seconds() / n_gaps
        if real_times is None:
            real_times = live_times

        super().__init__(tick_times=tick_times, live_times=live_times, real_times=real_times, name=name,
                         comment=comment, start_time=start_time)

        if energies is None:
            self.calibration = Calibration.channel_numbers_only(ticks.shape[1])
        elif isinstance(energies, Calibration):
            self.calibration = energies
        elif isinstance(energies, np.ndarray):
            self.calibration = Calibration.from_energies(energies, bin_label)
        self.ticks = ticks
        self.blank_rate = blank_rate
        self.msg_times = msg_times
        self._errors = errors

    @property
    def energies(self) -> np.ndarray:
        """
        Property which returns the lower energy bound foreach channel in the spectrum.

        Returns
        -------
        np.ndarray
        """
        return self.calibration.energies

    @property
    def n_channels(self) -> int:
        """
        Property which returns the number of channels in each spectrum per tick.

        Returns
        -------
        int
        """
        return self.calibration.n_channels

    @property
    def errors(self) -> np.ndarray:
        """
        Property which returns the uncertainties for each channel of each tick.

        Returns
        -------
        np.ndarray
        """
        if self._errors is None:
            assert (self.ticks >= 0.).all()
            self._errors = self.ticks**0.5
            self.log.debug("Using Poisson errors")
        return self._errors

    def get_single_tick(self, tick_number: int) -> Spectrum:
        """
        Method which retrieves a single tick.

        Parameters
        ----------
        tick_number : int
            Index of the tick to be retrieved.

        Returns
        -------
        Spectrum
        """
        return Spectrum(self.calibration, self.ticks[tick_number],
                        live_time=self.live_times[tick_number], real_time=self.real_times[tick_number],
                        error=self.errors[tick_number], name=self.name, comment=self.comment)

    def sum_ticks(self,
                  start: Union[datetime, None] = None,
                  end: Union[datetime, None] = None) -> Spectrum:
        """
        Method which returns a summed spectrum with optional start and end
        times.

        Parameters
        ----------
        start : datetime
            Start time for the spectrum sum. Optional.
        end : datetime
            End time for the spectrum sum. Optional.

        Returns
        -------
        Spectrum
        """
        t_mask = np.ones(self.tick_times.size, dtype=bool)
        if start is not None:
            t_mask = np.logical_and(t_mask, self.tick_times >= start)
        if end is not None:
            t_mask = np.logical_and(t_mask, self.tick_times < end)
        return Spectrum(self.calibration,
                        self.ticks[t_mask].sum(axis=0),
                        name=self.name,
                        live_time=self.live_times[t_mask].sum(),
                        real_time=self.real_times[t_mask].sum(),
                        error=np.sqrt(np.square(self.errors[t_mask]).sum(axis=0)),
                        comment=self.comment, start_time=self.start_time)

    def get_ticked_counts(self,
                          lower_lim: Union[int, float, None] = None,
                          upper_lim: Union[int, float, None] = None,
                          start: Union[datetime, None] = None,
                          end: Union[datetime, None] = None,
                          use_energy: bool = False) -> TickedCounts:
        """
        Method which sums channels in each tick's spectrum to make a TickedCounts instance.

        Parameters
        ----------
        lower_lim : int or float
            Optional channel (default) or energy for lower bound of the sum.
        upper_lim : int or float
            Optional channel (default) or energy for upper bound of the sum.
        start : datetime
            Start time for the spectrum sum. Optional.
        end : datetime
            End time for the spectrum sum. Optional.
        use_energy : bool
            Use energy values for the bounds, rather than channel number.

        Returns
        -------
        TickedCounts
        """
        c_mask = np.ones(self.n_channels, dtype=bool)
        t_mask = np.ones(self.tick_times.size, dtype=bool)

        if use_energy:
            if lower_lim is not None:
                c_mask = np.logical_and(c_mask, self.energies >= lower_lim)
            if upper_lim is not None:
                c_mask = np.logical_and(c_mask, self.energies < upper_lim)
        else:
            if lower_lim is not None:
                c_mask = np.logical_and(c_mask, np.arange(self.n_channels) >= lower_lim)
            if upper_lim is not None:
                c_mask = np.logical_and(c_mask, np.arange(self.n_channels) < upper_lim)

        if start is not None:
            t_mask = np.logical_and(t_mask, self.tick_times >= start)
        if end is not None:
            t_mask = np.logical_and(t_mask, self.tick_times < end)
        return TickedCounts(self.ticks[t_mask][:, c_mask].sum(axis=1),
                            self.tick_times[t_mask],
                            name=self.name,
                            live_times=self.live_times[t_mask],
                            real_times=self.real_times[t_mask],
                            msg_times=self.msg_times,
                            blank_rate=self.blank_rate,
                            errors=np.sqrt(np.square(self.errors).sum(axis=1)),
                            start_time=self.start_time)

    def subtract_background(self, background_spectrum: Spectrum) -> TickedSpectrum:
        """
        Function which takes a background spectrum and subtracts it,
        tick by tick, scaled by the correct live times.

        Parameters
        ----------
        background_spectrum : Spectrum
            Background spectrum to be subtracted.

        Returns
        -------
        TickedSpectrum
        """
        return self.simultaneous_subtraction(background_spectrum)

    def simple_windowed_spectra_iterations(self, tick_window_width: int = 1, mode: int = 0) -> int:
        """
        Method which calculates the number of iterations for a rolling window of a given length and mode.

        Parameters
        ----------
        tick_window_width : int
            Window width in whole ticks.
        mode : int
            Mode 0, 1, or 2

        Returns
        -------
        int
        """
        if mode == 0:
            return self.ticks.shape[0] - tick_window_width + 1
        elif mode == 1:
            return self.ticks.shape[0]
        elif mode == 2:
            return self.ticks.shape[0] + tick_window_width - 1

    def simple_windowed_spectra(self, tick_window_width: int = 1, mode: int = 0) -> Generator[Spectrum]:
        """
        Simple generator of variable length windowed spectra, equivalent to a rolling window across the time space

        Parameters
        ----------
        tick_window_width : int
            width of the window, in ticks. Default: 1.
        mode : int
            sets the mode of the rolling window. default: 0.
            0. windows are always full. For a TickedSpectrum with N ticks, there will be N - tick_window_width + 1
               spectra yielded.
            1. windows are always at least half full. For a TickedSpectrum with N ticks, there will be N spectra
               yielded.
            2. windows are always at least one tick full. For a TickedSpectrum with N ticks, there will be
               N + tick_window_width - 1 spectra yielded.
        """
        size = self.simple_windowed_spectra_iterations(tick_window_width, mode)
        if mode == 0:
            for i in range(size):
                yield Spectrum(self.calibration, self.ticks[i:i+tick_window_width].sum(axis=0), name=self.name,
                               error=np.sqrt(np.square(self.errors[i:i+tick_window_width]).sum(axis=0)),
                               comment=self.comment, live_time=self.live_times[i:i+tick_window_width].sum(),
                               real_time=self.real_times[i:i+tick_window_width].sum(),
                               start_time=self.tick_times[i])
        elif mode == 1:
            offset = -int(tick_window_width / 2)
            for i in range(offset, size + offset):
                yield Spectrum(self.calibration, self.ticks[max(i, 0):i+tick_window_width].sum(axis=0), name=self.name,
                               error=np.sqrt(np.square(self.errors[max(i, 0):i+tick_window_width]).sum(axis=0)),
                               comment=self.comment, live_time=self.live_times[max(i, 0):i+tick_window_width].sum(),
                               real_time=self.real_times[max(i, 0):i+tick_window_width].sum(),
                               start_time=self.tick_times[max(i, 0)])
        elif mode == 2:
            offset = -tick_window_width + 1
            for i in range(offset, size + offset):
                yield Spectrum(self.calibration, self.ticks[max(i, 0):i+tick_window_width].sum(axis=0), name=self.name,
                               error=np.sqrt(np.square(self.errors[max(i, 0):i+tick_window_width]).sum(axis=0)),
                               comment=self.comment, live_time=self.live_times[max(i, 0):i+tick_window_width].sum(),
                               real_time=self.real_times[max(i, 0):i+tick_window_width].sum(),
                               start_time=self.tick_times[max(i, 0)])
        yield None

    def resample(self, factor: float = 1.0) -> TickedSpectrum:
        """
        Method which uses the current spectrum (scaled by the 'scale' parameters) as the expectation for a new spectrum,
        sampled from either a normal or Poisson distribution.

        Parameters
        ----------
        factor : float
            Multiplication factor for the expectations of the new spectral counts.

        Returns
        -------
        new_ticked_spectrum : TickedSpectrum
            The resampled spectrum
        """
        new_ticked_spectrum = self.copy_me()
        if self._errors is None:
            new_ticked_spectrum.counts[self.ticks >= 0.] = np.random.poisson(self.ticks[self.ticks >= 0.] * factor)
        else:
            errors = np.sqrt(self.errors**2*factor)
            new_ticked_spectrum.counts = np.random.normal(self.ticks * factor, errors)
            new_ticked_spectrum.counts[new_ticked_spectrum.counts < 0.] = 0.
            new_ticked_spectrum._errors = errors
        return new_ticked_spectrum

    def consecutive_addition(self, other: TickedSpectrum) -> TickedSpectrum:
        """
        Method which consecutively sums another TickedSpectrum instance to this one.

        Parameters
        ----------
        other : TickedSpectrum
            TickedSpectrum instance to sum to this one.

        Returns
        -------
        TickedSpectrum
        """
        if self.calibration != other.calibration:
            other = other.rebin(self.calibration)
        return TickedSpectrum(energies=self.energies,
                              ticks=np.concatenate((self.ticks, other.ticks)),
                              tick_times=np.concatenate((self.tick_times, other.tick_times)),
                              live_times=np.concatenate((self.live_times, other.live_times)),
                              real_times=np.concatenate((self.real_times, other.real_times)),
                              name=self.name,
                              msg_times=None if self.msg_times is None or other.msg_times is None else np.concatenate(
                                  (self.msg_times, other.msg_times)
                              ),
                              blank_rate=(None if self.blank_rate is None or other.blank_rate is None else
                                          np.concatenate((self.blank_rate, other.blank_rate))),
                              errors=None if self._errors is None and other._errors is None else np.concatenate(
                                  (self.errors, other.errors)
                              ),
                              start_time=self.start_time,
                              comment=self.comment)

    def _simultaneous_addition_spectrum(self, spectrum: Spectrum) -> TickedSpectrum:
        if spectrum.calibration != self.calibration:
            spectrum = spectrum.rebin(self.calibration)
        scales = (self.live_times / spectrum.live_time)[:, np.newaxis]
        return TickedSpectrum(
            self.calibration,
            self.tick_times,
            self.ticks + spectrum.counts * scales,
            live_times=self.live_times,
            real_times=self.real_times,
            errors=np.sqrt(np.square(self.errors) + np.square(spectrum.error * scales)),
            name=self.name,
            comment=self.comment,
            start_time=self.start_time,
            msg_times=self.msg_times,
            blank_rate=self.blank_rate
        )

    def _simultaneous_addition_ticked_spectrum(self, other: TickedSpectrum) -> TickedSpectrum:
        if self.calibration != other.calibration:
            other = other.rebin(self.calibration)
        if self.tick_times.size != other.tick_times.size:
            self.log.warning(f'{self.name} has {self.tick_times.size} ticks, but {other.name} has '
                             f'{self.tick_times.size} ticks. The longer sequence will be shortened.')
        n_ticks = min(self.tick_times.size, other.tick_times.size)
        scales = (self.live_times[:n_ticks] / other.live_times[:n_ticks])[:, np.newaxis]
        return TickedSpectrum(
            self.calibration,
            self.tick_times,
            self.ticks[:n_ticks] + other.ticks[:n_ticks] * scales,
            live_times=self.live_times[:n_ticks],
            real_times=self.real_times[:n_ticks],
            errors=np.sqrt(np.square(self.errors[:n_ticks]) + np.square(other.errors[:n_ticks]) * scales),
            name=self.name,
            comment=self.comment,
            start_time=self.start_time,
            msg_times=None if self.msg_times is None else self.msg_times[:n_ticks],
            blank_rate=None if self.blank_rate is None else self.blank_rate[:n_ticks]
        )

    def _simultaneous_subtraction_spectrum(self, spectrum: Spectrum) -> TickedSpectrum:
        if spectrum.calibration != self.calibration:
            spectrum = spectrum.rebin(self.calibration)
        scales = (self.live_times / spectrum.live_time)[:, np.newaxis]
        return TickedSpectrum(
            self.calibration,
            self.tick_times,
            self.ticks - spectrum.counts * scales,
            live_times=self.live_times,
            real_times=self.real_times,
            errors=np.sqrt(np.square(self.errors) + np.square(spectrum.error * scales)),
            name=self.name,
            comment=self.comment,
            start_time=self.start_time,
            msg_times=self.msg_times,
            blank_rate=self.blank_rate
        )

    def _simultaneous_subtraction_ticked_spectrum(self, other: TickedSpectrum) -> TickedSpectrum:
        if self.calibration != other.calibration:
            other = other.rebin(self.calibration)
        if self.tick_times.size != other.tick_times.size:
            self.log.warning(f'{self.name} has {self.tick_times.size} ticks, but {other.name} has '
                             f'{self.tick_times.size} ticks. The longer sequence will be shortened.')
        n_ticks = min(self.tick_times.size, other.tick_times.size)
        scales = (self.live_times[:n_ticks] / other.live_times[:n_ticks])[:, np.newaxis]
        return TickedSpectrum(
            self.calibration,
            self.tick_times,
            self.ticks[:n_ticks] - other.ticks[:n_ticks] * scales,
            live_times=self.live_times[:n_ticks],
            real_times=self.real_times[:n_ticks],
            errors=np.sqrt(np.square(self.errors[:n_ticks]) + np.square(other.errors[:n_ticks]) * scales),
            name=self.name,
            comment=self.comment,
            start_time=self.start_time,
            msg_times=None if self.msg_times is None else self.msg_times[:n_ticks],
            blank_rate=None if self.blank_rate is None else self.blank_rate[:n_ticks]
        )

    def simultaneous_addition(self, other: Union[TickedSpectrum, Spectrum]) -> TickedSpectrum:
        """
        Method which simultaneously sums a Spectrum or another TickedSpectrum to this TickedSpectrum.

        Parameters
        ----------
        other : Spectrum or TickedSpectrum

        Returns
        -------
        TickedSpectrum
        """
        if isinstance(other, TickedSpectrum):
            return self._simultaneous_addition_ticked_spectrum(other)
        else:
            return self._simultaneous_addition_spectrum(other)

    def simultaneous_subtraction(self, other: Union[TickedSpectrum, Spectrum]) -> TickedSpectrum:
        """
        Method which simultaneously subtracts a Spectrum or another TickedSpectrum from this TickedSpectrum.

        Parameters
        ----------
        other : Spectrum or TickedSpectrum

        Returns
        -------
        TickedSpectrum
        """
        if isinstance(other, TickedSpectrum):
            return self._simultaneous_subtraction_ticked_spectrum(other)
        else:
            return self._simultaneous_subtraction_spectrum(other)

    def rebin(self, new_calibration: Calibration) -> TickedSpectrum:
        """
        Function which rebins spectrum with new calibration coefficients and number of channels.

        Parameters
        ----------
        new_calibration : Calibration
            Calibration of required rebinned spectrum.

        Returns
        -------
        new_spectrum : Spectrum
            Rebinned spectrum with calibration equal to 'new_calibration'.
        """

        # Get upper and lower energy values for each current channel
        energies_upper = self.calibration.upper_bin_limits
        energies_lower = self.calibration.lower_bin_limits

        # Get upper and lower energy values for each new channel
        new_energies_upper = new_calibration.upper_bin_limits
        new_energies_lower = new_calibration.lower_bin_limits

        # Generate empty array for new counts
        new_ticked_spectrum = self.copy_me()
        new_ticked_spectrum.calibration = new_calibration
        new_ticked_spectrum.ticks = np.zeros((new_calibration.n_channels, self.tick_times.size))
        new_ticked_spectrum._errors = None if self._errors is None else np.zeros((new_calibration.n_channels,
                                                                                  self.tick_times.size))

        # Check there will be counts in the new spectrum
        assert self.energies[0] < new_energies_upper[-1] and self.energies[-1] > new_energies_upper[0]

        zeros = np.zeros(new_calibration.n_channels)
        # Distribute the counts in each channel to the new spectrum
        for channel in range(self.n_channels):
            w_max = energies_upper[channel]
            w_min = energies_lower[channel]

            energy_overlap = np.maximum(np.minimum(new_energies_upper, w_max) -
                                        np.maximum(new_energies_lower, w_min),
                                        zeros)[np.newaxis, :]
            new_ticked_spectrum.ticks += self.ticks[:, channel] * energy_overlap / (w_max - w_min)
            if self._errors is not None:
                new_ticked_spectrum._errors += self._errors[channel]**2 * energy_overlap / (w_max - w_min)

        if self._errors is not None:
            new_ticked_spectrum._errors = np.sqrt(new_ticked_spectrum._errors)

        # Return the new spectrum
        return new_ticked_spectrum


class TickedSpectrumArray(TickedArrayBase):
    """
    Array of TickedSpectrum instances.

    Attributes
    ----------
    elements : OrderedDict[str, TickedSpectrum]
        OrderedDict, mapping measurements by their names.

    Methods
    -------
    tick_times() -> Array[datetime]
        Property which returns a list of datetime instances labelling the time at which each tick occurred.
    n_ticks() -> int
        Property which returns the number of ticks in the first member of the array.
    names() -> List[str]
        Property which returns a list of the measurement names which are keys to the mapping.
    number_of_members() -> int
        Property which returns the number of members of the array.
    start_times() -> List[datetime]
        Property which returns a list of datetime instances which give the times at which each member of the array
        began to be collected.
    contains_all(names: List[str]) -> bool
        Method which returns True if all the strings in the argument match the names of elements in the array.
    names_match(other: ArrayBase) -> bool
        Method which returns True if all the names of the other array are present in this array, and vice-versa.
    consecutive_addition(other: TickedSpectrumArray) -> TickedSpectrumArray
        Method which returns the consecutive sum of the TickedSpectrum with another Array.
    simultaneous_addition(other: TickedSpectrumArray) -> TickedSpectrumArray
        Method which returns the simultaneous sum of the TickedSpectrum with another Array.
    simultaneous_subtraction(other: Union[SpectrumArray, TickedSpectrumArray]) -> TickedSpectrumArray:
        Method which returns the simultaneous subtraction of another CountsArray from the Array.
    subtract_backgrounds(self, backgrounds: SpectrumArray) -> TickedSpectrumArray:
        Method which subtracts the background from each member of the array, using the array of background counts
        supplied as the argument.
    simultaneous_sum() -> TickedSpectrum
        Method which simultaneously sums each member of the array into a single TickedSpectrum instance.
    consecutive_sum() -> TickedSpectrum
        Must be implemented by subclasses.
    resample(factor: float) -> TickedSpectrumArray
        Method which returns a new array with each member resampled after scaling by the factor parameter.
    masked_array(names: List[str]) -> TickedSpectrumArray
        Method which returns a new array consisting only of members with the names in the given list of names.
    get_single_tick(tick_num: int) -> SpectrumArray
        Method which retrieves a SpectrumArray of Spectra for a single tick of the TickedSpectrumArray.
    sum_ticks(start: datetime, end: datetime) -> SpectrumArray
        Must be implemented by subclasses.
    get_ticked_counts_array() -> TickedCountsArray
        Method which sums channels  with optional channel or energy limits, into a TickedCountsArray with optional
        timing limits.
    simple_windowed_spectrum_arrays_iterations(tick_window_width: int = 1, mode: int = 0) -> int
        Method which calculates the number of iterations for a rolling window of a given length and mode.
    simple_windowed_spectrum_arrays(tick_window_width: int = 1, mode: int = 0) -> Iterator[SpectrumArray]
        Simple generator of variable length windowed SpectrumArrays, equivalent to a rolling window across the time
        space.
    sum_ticks(start: Union[datetime, None] = None, end: Union[datetime, None] = None) -> SpectrumArray
        Sum the ticks within optional time limits.
    consecutive_sum(self) -> TickedSpectrum
        Method which consecutively sums the members of the array.
    """
    def __init__(self, t_spectra: List[TickedSpectrum]):
        super(TickedSpectrumArray, self).__init__(t_spectra)

    def __getitem__(self, item) -> TickedSpectrum:
        return super(TickedSpectrumArray, self).__getitem__(item)

    def get_single_tick(self, tick_num: int) -> SpectrumArray:
        """
        Method which retrieves a SpectrumArray of Spectra for a single tick of the TickedSpectrumArray.

        Parameters
        ----------
        tick_num : int
            Index of the tick to retrieve

        Returns
        -------
        SpectrumArray
        """
        return SpectrumArray([ts.get_single_tick(tick_num) for ts in self.elements.values()])

    def simultaneous_sum(self) -> TickedSpectrum:
        """
        Method which simultaneously sums each member of the array into a single TickedSpectrum instance.

        Returns
        -------
        TickedSpectrum
        """
        result = None
        for ticked_spectrum in self:
            if result is None:
                result = ticked_spectrum
            else:
                result = result.simultaneous_addition(ticked_spectrum)
        return result

    def get_ticked_counts_array(self,
                                lower_lim: Union[int, float, None] = None,
                                upper_lim: Union[int, float, None] = None,
                                start: Union[datetime, None] = None,
                                end: Union[datetime, None] = None,
                                use_energy: bool = False) -> TickedCountsArray:
        """
        Method which sums channels  with optional channel or energy limits, into a TickedCountsArray with optional
        timing limits.

        Parameters
        ----------
        lower_lim : Union[int, float]
            Optional lower channel (default) or energy limit for the sum.
        upper_lim : Union[int, float]
            Optional upper channel (default) or energy limit for the sum.
        start : datetime
            optional time defining the start of the TickedCountsArray
        end : datetime
            optional time defining the end of the TickedCountsArray
        use_energy : bool
            Sets the limits to be energy values rather than channel numbers.

        Returns
        -------
        TickedCountsArray
        """
        return TickedCountsArray([self[det].get_ticked_counts(lower_lim, upper_lim, start, end, use_energy)
                                  for det in self.names])

    def simple_windowed_spectrum_arrays_iterations(self, tick_window_width: int = 1, mode: int = 0) -> int:
        """
        Method which calculates the number of iterations for a rolling window of a given length and mode.

        Parameters
        ----------
        tick_window_width : int
            Window width in whole ticks.
        mode : int
            Mode 0, 1, or 2

        Returns
        -------
        int
        """
        return self[self.names[0]].simple_windowed_spectra_iterations(tick_window_width, mode)

    def simple_windowed_spectrum_arrays(self, tick_window_width: int = 1, mode: int = 0) -> Iterator[SpectrumArray]:
        """
        Simple generator of variable length windowed SpectrumArrays, equivalent to a rolling window across the time
        space.

        Parameters
        ----------
        tick_window_width : int
            width of the window, in ticks. Default: 1.
        mode : int
            sets the mode of the rolling window. default: 0.
            Modes:
                0 - windows are always full. For a TickedSpectrum with N ticks, there will be N - tick_window_width + 1
                    spectra yielded.
                    |====================================================| Ticks
                    |====|----------------------------------------->|====| Windows
                1 - windows are always at least half full. For a TickedSpectrum with N ticks, there will be N spectra
                    yielded.
                      |====================================================| Ticks
                    |====|--------------------------------------------->|====| Windows
                2 - windows are always at least one tick full. For a TickedSpectrum with N ticks, there will be
                    N + tick_window_width - 1 spectra yielded.
                         |====================================================| Ticks
                    |====|--------------------------------------------------->|====| Windows
        """
        generators = [self[d].simple_windowed_spectra(tick_window_width, mode) for d in self.names]
        while 1:
            spectra = [next(gen) for gen in generators]
            if spectra[0] is None:
                break
            yield SpectrumArray(*spectra)

    def sum_ticks(self, start: Union[datetime, None] = None, end: Union[datetime, None] = None) -> SpectrumArray:
        """
        Sum the ticks within optional time limits.

        Parameters
        ----------
        start : datetime
            Optional start time for the sum.
        end : datetime
            Optional end time for the sum.

        Returns
        -------
        SpectrumArray
        """
        return SpectrumArray([self[det].sum_ticks(start, end) for det in self.names])

    def consecutive_sum(self) -> TickedSpectrum:
        """
        Method which consecutively sums the members of the array.

        Returns
        -------
        TickedSpectrum
        """
        result = None
        for ticked_spectrum in self:
            if result is None:
                result = ticked_spectrum
            else:
                result = result.consecutive_addition(ticked_spectrum)
        return result



import numpy as np
from tqdm import trange


class DeconvolutionAlgorithm:
    def __init__(self, response_matrix: np.ndarray):
        self.response_matrix = response_matrix

    def deconvolve_measurement(
            self,
            measurement: np.ndarray,
            iterations: int = 1000,
            init: np.ndarray = None
    ) -> [np.ndarray, np.ndarray]:
        raise NotImplementedError

    def bootstrap_deconvolve_measurement(
            self,
            measurement: np.ndarray,
            errors: np.ndarray,
            iterations: int = 1000,
            bootstraps: int = 100,
            init: np.ndarray = None,
            progress: bool = False
    ) -> [np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        decons = np.zeros((bootstraps, self.response_matrix.shape[1]))
        recons = np.zeros((bootstraps, self.response_matrix.shape[0]))

        rng = trange if progress else range

        for i in rng(bootstraps):
            new_measurement = np.random.normal(measurement, errors)
            decons[i, :], recons[i, :] = self.deconvolve_measurement(new_measurement, iterations, init)

        return decons.mean(axis=0), recons.mean(axis=0), decons.std(axis=0), recons.std(axis=0)


class MLEM(DeconvolutionAlgorithm):
    def deconvolve_measurement(
            self,
            measurement: np.ndarray,
            iterations: int = 1000,
            init: np.ndarray = None
    ) -> [np.ndarray, np.ndarray]:
        decon = np.ones(self.response_matrix.shape[1]) if init is None else init
        weight = self.response_matrix.transpose() * measurement
        norm = self.response_matrix.sum(axis=0)
        norm[norm == 0.0] = 1.0
        while iterations > 0:
            ir = np.dot(self.response_matrix, decon)
            ir[ir < 1.e-10] = 1.e-10
            decon = decon * np.dot(weight, 1.0 / ir) / norm
            iterations -= 1

        recon = np.dot(self.response_matrix, decon)

        return decon, recon



from __future__ import annotations
import numpy as np
from datetime import datetime
from typing import Union, List, Iterator, Tuple, Any
import matplotlib.pyplot as plt

from geiger.core.base_classes import MeasurementBase, TickedMeasurement, ArrayBase, TickedArrayBase
from geiger.utilities.plotting import TickedCountsPlotter


class Counts(MeasurementBase):
    """
    Class which stores counts, live and real tick_times and errors for count data.

    Parameters
    ----------
    counts : float
        Number of counts measured within live_time by the detector.
    live_time : float
        Live_time of detector during acquisition.
    real_time : float
        Real_time of detector during acquisition.
    name : str
        Label for the count data.
    comment : str
        Comment to attach to the count data.
    start_time : datetime or None
        Time at which the acquisition started.

    Methods
    -------
    consecutive_addition(other: Counts) -> Counts
        Method which subtracts Counts data from the current Counts data, equivalent to subtracting a subsection of a
        measurement.
    """
    def __init__(self,
                 counts: float,
                 live_time: float,
                 real_time: float,
                 name: str = '',
                 error: Union[float, None] = None,
                 comment: str = '',
                 start_time: Union[datetime, None] = None):
        self.counts = counts
        self._error = error
        super(Counts, self).__init__(live_time=live_time,
                                     real_time=real_time,
                                     name=name,
                                     comment=comment,
                                     start_time=start_time)

    @property
    def error(self) -> float:
        """
        Property which returns the uncertainty in the count data.

        Returns
        -------
        float
        """
        if self._error is None:
            return self.counts**0.5
        else:
            return self._error

    @property
    def count_rate(self) -> float:
        """
        Property which returns the count rate of the measurement. I.e. counts / live_time

        Returns
        -------
        float
        """
        return (self.counts / self.live_time) if self.live_time > 0. else 0.

    def resample(self, scale: float = 1.0) -> Counts:
        """
        Method which returns a new Counts instance with the counts data resampled from a normal distribution from the
        counts and error data, scaled by the 'scale' factor.

        Parameters
        ----------
        scale : float
            Scale factor by which the counts data will be multiplied before resampling.

        Returns
        -------
        Counts
        """
        new_counts = np.random.normal(self.counts*scale, np.sqrt(scale*self.error**2))
        return Counts(new_counts, live_time=self.live_time, real_time=self.real_time, name=self.name,
                      error=np.sqrt(scale*self.error**2), comment=self.comment)

    def subtract_background(self, bg_counts: Counts) -> Counts:
        """
        Function which takes a background Counts object and subtracts it,
        scaled by the correct live time, returning the resultant Counts.

        Parameters
        ----------
        bg_counts : Counts
            Background counts to be subtracted.

        Returns
        -------
        Counts
        """
        return super(Counts, self).subtract_background(bg_counts)

    def consecutive_addition(self, other: Counts) -> Counts:
        """
        Method which adds Counts data to the current Counts data, equivalent to adding two consecutive measurements
        together, adding their counts and live and real times together.

        Parameters
        ----------
        other : Counts
            Counts to add.

        Returns
        -------
        Counts
        """
        counts = Counts(self.counts + other.counts,
                        live_time=self.live_time + other.live_time,
                        real_time=self.real_time + other.real_time,
                        error=(self.error**2 + other.error**2)**0.5,
                        name=self.name)
        return counts

    def simultaneous_addition(self, other: Counts) -> Counts:
        """
        Method which adds Counts data to the current Counts data, equivalent to adding two simultaneous measurements
        together, adding their counts together, scaling 'other' counts to have equal live time.

        Parameters
        ----------
        other : Counts
            Counts to add.

        Returns
        -------
        Counts
        """
        scale = self.live_time / other.live_time
        counts = Counts(self.counts + other.counts * scale,
                        live_time=self.live_time,
                        real_time=self.real_time,
                        error=(self.error**2 + (other.error*scale)**2)**0.5,
                        name=self.name)
        return counts

    def consecutive_subtraction(self, other: Counts) -> Counts:
        """
        Method which subtracts Counts data from the current Counts data, equivalent to subtracting a subsection of a
        measurement.

        Parameters
        ----------
        other : Counts
            Counts to subtract.

        Returns
        -------
        Counts
        """
        counts = Counts(self.counts - other.counts,
                        live_time=self.live_time - other.live_time,
                        real_time=self.real_time - other.real_time,
                        error=(self.error**2 + other.error**2)**0.5,
                        name=self.name)
        return counts

    def simultaneous_subtraction(self, other: Counts) -> Counts:
        """
        Method which subtracts Counts data from the current Counts data, equivalent to subtracting a signal present
        throughout the measurement, e.g. the background.

        Parameters
        ----------
        other : Counts
            Counts to subtract.

        Returns
        -------
        Counts
        """
        scale = self.live_time / other.live_time
        counts = Counts(self.counts - other.counts * scale,
                        live_time=self.live_time,
                        real_time=self.real_time,
                        error=(self.error**2 + (other.error*scale)**2)**0.5,
                        name=self.name)
        return counts

    def __str__(self) -> str:
        return (f"<Counts counts='{self.counts:.1f}' live_time='{self.live_time:.1f}' real_time='{self.real_time:.1f}' "
                f"error='{self.error:.1f}' name='{self.name}'  comment='{self.comment}' />")

    def __eq__(self, other: Counts) -> bool:
        return self.counts == other.counts and self.live_time == other.live_time


class CountsArray(ArrayBase):
    """
    Array of Counts objects, indexed by their 'name' attribute.

    Parameters
    ----------
    counts : List[Counts]
        List of Counts instances. These will become the members of the array.

    Methods
    -------
    consecutive_addition(other: Counts) -> Counts
        Method which subtracts Counts data from the current Counts data, equivalent to subtracting a subsection of a
        measurement.

    Notes
    -----
    Users will probably not need to create CountsArray instances, they will be used after being created inside the
    SMLReader.
    """

    def __init__(self, counts: List[Counts]):
        super(CountsArray, self).__init__(counts)

    def __getitem__(self, item) -> Counts:
        return super(CountsArray, self).__getitem__(item)

    def get(self, item: str, default_val: Any = Counts(0., 0., 0.)) -> Counts:
        return super(CountsArray, self).get(item, default_val)

    def simultaneous_sum(self) -> Counts:
        """
        Method which returns the simultaneous sum of all of the members of the array.

        Returns
        -------
        Counts
        """
        c_sum = Counts(counts=0.,
                       live_time=float(np.mean([c.live_time for c in self.elements.values()])),
                       real_time=float(np.mean([c.real_time for c in self.elements.values()])),
                       name=f'Simultaneously summed counts from ' + ', '.join(self.names))
        for c in self.elements.values():
            c_sum = c_sum.simultaneous_addition(c)
        return c_sum

    def consecutive_sum(self) -> Counts:
        """
        Method which returns the consecutive sum of all of the members of the array.

        Returns
        -------
        Counts
        """
        c_sum = Counts(counts=0., live_time=0., real_time=0.,
                       name=f'Consecutively summed counts from ' + ', '.join(self.names))
        for c in self.elements.values():
            c_sum = c_sum.consecutive_addition(c)
        return c_sum

    def subtract_backgrounds(self, backgrounds: CountsArray) -> CountsArray:
        """
        Method which subtracts the background from each member of the array, using the array of background counts
        supplied as the argument.

        Parameters
        ----------
        backgrounds : CountsArray

        Returns
        -------
        CountsArray
        """
        return super(CountsArray, self).subtract_backgrounds(backgrounds)

    def consecutive_addition(self, other: CountsArray) -> CountsArray:
        """
        Method which returns the consecutive sum of the CountsArray with another Array.

        Parameters
        ----------
        other : CountsArray
            The Array instance to add to this one.

        Returns
        -------
        CountsArray
        """
        return super(CountsArray, self).consecutive_addition(other)

    def consecutive_subtraction(self, other: CountsArray) -> CountsArray:
        """
        Method which returns the consecutive subtraction of another CountsArray from the Array.

        Parameters
        ----------
        other : Array
            The Array instance to subtract from this one.

        Returns
        -------
        CountsArray
        """
        return super(CountsArray, self).consecutive_subtraction(other)

    def simultaneous_addition(self, other: CountsArray) -> CountsArray:
        """
        Method which returns the simultaneous sum of the CountsArray with another Array.

        Parameters
        ----------
        other : Array
            The Array instance to add to this one.

        Returns
        -------
        CountsArray
        """
        return super(CountsArray, self).simultaneous_addition(other)

    def simultaneous_subtraction(self, other: CountsArray) -> CountsArray:
        """
        Method which returns the simultaneous subtraction of another CountsArray from the Array.

        Parameters
        ----------
        other : Array
            The Array instance to subtract from this one.

        Returns
        -------
        CountsArray
        """
        return super(CountsArray, self).simultaneous_subtraction(other)

    def resample(self, factor: float) -> CountsArray:
        """
        Method which returns a new array with each member resampled after scaling by the factor parameter.

        Parameters
        ----------
        factor : float
            Scale by which the measurements will be scaled before resampling.

        Returns
        -------
        CountsArray
        """
        return super(CountsArray, self).resample(factor)

    def masked_array(self, names: List[str]) -> CountsArray:
        """
        Method which returns a new array consisting only of members with the names in the given list of names.

        Parameters
        ----------
        names : List[str]

        Returns
        -------
        Array
        """
        return super(CountsArray, self).masked_array(names)

    def __str__(self):
        return f"Counts : " + ", ".join([f"{det}:{self[det].counts:.1f}" for det in self.elements])


class TickedCounts(TickedMeasurement):
    """
    Class which contains details about the counts for multiple ticks (sequential short periods of time).

    Attributes
    ----------
    ticks : np.ndarray
        Array of counts for each tick.
    tick_times : np.ndarray
        Array of datetime instances, marking the time at which each tick ended or began.
    live_times : np.ndarray
        Array of floats, labelling the live time for each tick.
    real_times : np.ndarray
        Array of floats, labelling the real time for each tick.
    msg_times : np.ndarray
        Possible array of datetime instances, marking the times at which the tick messages arrived.
    blank_rate : np.ndarray
        Possible array of floats, marking the rate of blanking pulses measured during each tick.
    name : str
        Label for the count data.
    comment : str
        Comment to attach to the count data.
    start_time : datetime or None
        Time at which the acquisition started.

    Methods
    -------
    errors() -> np.ndarray
        Property which returns the uncertainties for each tick.
    get_single_tick(tick_num: int) -> Counts
        Method which returns a single tick's worth of counts, with the appropriate live time and real time.
    sum_ticks(start: datetime = None, end: datetime = None) -> Counts
        Method which returns a Counts instance containing the summed counts for all ticks, or ticks in the range
        start:end.
    resample(scale: float = 1.0) -> TickedCounts
        Method which returns a resampled TickedCounts instance.
    plot(**kwargs) -> Tuple[plt.Figure, plt.Axes]
        Simple generator of variable length windowed spectra, equivalent to a rolling window across the time space.
    simple_windowed_counts(tick_window_width: int = 1, mode: int = 0) -> Iterator[Counts]
        Simple generator of variable length windowed spectra, equivalent to a rolling window across the time space.
    subtract_background(bg_counts: Counts) -> TickedCounts
        Method which subtracts a background signal from each of the ticks, using simultaneous subtraction.
    """
    def __init__(self,
                 ticks: np.ndarray,
                 tick_times: np.ndarray,
                 live_times: np.ndarray,
                 real_times: np.ndarray,
                 name: str = '',
                 msg_times: Union[np.ndarray, None] = None,
                 blank_rate: Union[np.ndarray, None] = None,
                 errors: Union[np.ndarray, None] = None,
                 start_time: Union[np.ndarray, None] = None,
                 comment: str = ''):
        """
        Parameters
        ----------
        ticks : np.ndarray
            Array of counts for each tick.
        tick_times : np.ndarray
            Array of datetime instances, marking the time at which each tick ended or began.
        name : str
            Label for the count data.
        live_times : np.ndarray
            Array of floats, labelling the live time for each tick.
        real_times : np.ndarray
            Array of floats, labelling the real time for each tick.
        msg_times : np.ndarray
            Optional array of datetime instances, marking the times at which the tick messages arrived.
        blank_rate : np.ndarray
            Optional array of floats, marking the rate of blanking pulses measured during each tick.
        errors : np.ndarray
            Optional array of floats, marking the uncertainties of the counts values. If None, defaults to Poisson
            uncertainties.
        start_time : datetime
            Time at which the measurement began.
        comment : str
            Optional string comment to attach to the measurement.
        """

        super().__init__(tick_times=tick_times, live_times=live_times, real_times=real_times, name=name,
                         comment=comment, start_time=start_time)

        self.ticks = ticks
        self.msg_times = msg_times
        self.blank_rate = blank_rate
        self._errors = errors

    @property
    def errors(self) -> np.ndarray:
        """
        Property which returns the uncertainties for each tick.

        Returns
        -------
        np.ndarray
        """
        if self._errors is None:
            return np.abs(self.ticks) ** 0.5
        else:
            return self._errors

    def get_single_tick(self, tick_num: int) -> Counts:
        """
        Method which returns a single tick's worth of counts, with the appropriate live time and real time.

        Parameters
        ----------
        tick_num : int
            The index for the tick to retrieve.

        Returns
        -------
        Counts
        """
        return Counts(counts=self.ticks[tick_num],
                      live_time=self.live_times[tick_num],
                      real_time=self.real_times[tick_num],
                      error=self.errors[tick_num],
                      name=self.name + (' ' if len(self.name) > 0 else '') + f'Tick {tick_num}',
                      comment=self.comment)

    def sum_ticks(self, start: Union[datetime, None] = None, end: Union[datetime, None] = None) -> Counts:
        """
        Method which returns a Counts instance containing the summed counts for all ticks, or ticks in the range
        start:end.

        Parameters
        ----------
        start : Union[datetime, None]
            The start time from which to sum the ticks. If None, sums from the first tick in the sequence.
        end : Union[datetime, None]
            The end time up to which to sum the ticks. If None, sums up to the last tick in the sequence.

        Returns
        -------
        Counts
        """
        t_mask = np.ones(self.tick_times.size, dtype=bool)
        if start is not None:
            t_mask = np.logical_and(t_mask, self.tick_times >= start)
        if end is not None:
            t_mask = np.logical_and(t_mask, self.tick_times < end)
        counts = Counts(self.ticks[t_mask].sum(axis=0),
                        live_time=self.live_times[t_mask].sum(axis=0),
                        real_time=self.real_times[t_mask].sum(axis=0),
                        name=self.name, comment=self.comment,
                        error=None if self._errors is None else np.sqrt(np.square(self.errors[t_mask]).sum()),
                        start_time=self.start_time if start is None else start)
        return counts

    def resample(self, scale: float = 1.0) -> TickedCounts:
        """
        Method which returns a resampled TickedCounts instance.

        Parameters
        ----------
        scale : float
            Factor by which to multiply the TickedCounts instance before resampling.

        Returns
        -------
        TickedCounts
        """
        new_ticks = np.random.poisson(self.ticks*scale)
        return TickedCounts(new_ticks, self.tick_times, name=self.name, live_times=self.live_times*scale,
                            real_times=self.real_times, msg_times=self.msg_times, blank_rate=self.blank_rate,
                            errors=self.errors, start_time=self.start_time)

    def plot(self, other: List[TickedCounts] = None, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """
        Method which plots the TickedCounts as a function of time, returning the figure and axes.

        Parameters
        ----------
        other: List[TickedCounts]
            Other TickedCounts to plot together

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
        """
        if other is None:
            other = []
        return TickedCountsPlotter(ticked_counts=[self, *other], **kwargs).plot()

    def simple_windowed_counts(self, tick_window_width: int = 1, mode: int = 0) -> Iterator[Counts]:
        """
        Simple generator of variable length windowed spectra, equivalent to a rolling window across the time space.
        Modes:
        0 - windows are always full. For a TickedSpectrum with N ticks, there will be N - tick_window_width + 1
        spectra yielded.
        1 - windows are always at least half full. For a TickedSpectrum with N ticks, there will be N spectra
        yielded.
        2 - windows are always at least one tick full. For a TickedSpectrum with N ticks, there will be
        N + tick_window_width - 1 spectra yielded.

        Parameters
        ----------
        tick_window_width : int
            width of the window, in ticks. Default: 1.
        mode : int
            sets the mode of the rolling window. default: 0.

        Returns
        -------
        Iterator[Counts]
        """
        if mode == 0:
            for i in range(self.ticks.shape[0] - tick_window_width + 1):
                yield Counts(self.ticks[i:i+tick_window_width].sum(axis=0), name=self.name,
                             error=np.sqrt(np.square(self.errors[i:i+tick_window_width]).sum(axis=0)),
                             comment=self.comment, live_time=self.live_times[i:i+tick_window_width].sum(),
                             real_time=self.real_times[i:i+tick_window_width].sum(),
                             start_time=self.tick_times[i])
        elif mode == 1:
            for i in range(-int(tick_window_width / 2), self.ticks.shape[0] - int(tick_window_width / 2)):
                yield Counts(self.ticks[max(i, 0):i+tick_window_width].sum(axis=0), name=self.name,
                             error=np.sqrt(np.square(self.errors[max(i, 0):i+tick_window_width]).sum(axis=0)),
                             comment=self.comment, live_time=self.live_times[max(i, 0):i+tick_window_width].sum(),
                             real_time=self.real_times[max(i, 0):i+tick_window_width].sum(),
                             start_time=self.tick_times[max(i, 0)])
        elif mode == 2:
            for i in range(-tick_window_width+1, self.ticks.shape[0]):
                yield Counts(self.ticks[max(i, 0):i+tick_window_width].sum(axis=0), name=self.name,
                             error=np.sqrt(np.square(self.errors[max(i, 0):i+tick_window_width]).sum(axis=0)),
                             comment=self.comment, live_time=self.live_times[max(i, 0):i+tick_window_width].sum(),
                             real_time=self.real_times[max(i, 0):i+tick_window_width].sum(),
                             start_time=self.tick_times[max(i, 0)])

    def subtract_background(self, bg_counts: Counts) -> TickedCounts:
        """
        Method which subtracts a background signal from each of the ticks, using simultaneous subtraction.

        Note: Since ticks are short (~0.2s) the noise in the counts is often of the order of the counts themselves, and
        hence the background subtracted counts are often negative.

        Parameters
        ----------
        bg_counts : Counts
            Counts instance containing information about the background count rate.

        Returns
        -------
        TickedCounts
        """
        return self.simultaneous_subtraction(bg_counts)

    def _simultaneous_op_counts(self, other: Counts, add: bool):
        """
        Not sure why you'd want to do this, but it makes logical sense, so here it is.

        Parameters
        ----------
        other : Counts

        Returns
        -------
        TickedCounts
        """
        scale = self.live_times / other.live_time
        return TickedCounts(ticks=self.ticks + (1. if add else -1.)*other.counts * scale,
                            tick_times=self.tick_times,
                            name=self.name,
                            live_times=self.live_times,
                            real_times=self.real_times,
                            msg_times=self.msg_times,
                            blank_rate=self.blank_rate,
                            errors=np.sqrt(self.errors ** 2 + other.error ** 2 * scale),
                            start_time=self.start_time)

    def _simultaneous_op_ticked_counts(self, other: TickedCounts, add: bool):
        """
        Simultaneous sum of two CountsArrays.

        Parameters
        ----------
        other : TickedCounts

        Returns
        -------
        TickedCounts
        """
        if self.n_ticks != other.n_ticks:
            self.log.warning(f'Simultaneous Sum: {self.name} has {self.n_ticks} ticks, but {other.name} has '
                             f'{other.n_ticks} ticks. The longer measurement will be shortened.')
        n_ticks = min(self.n_ticks, other.n_ticks)
        scale = self.live_times[:n_ticks] / other.live_times[:n_ticks]
        return TickedCounts(ticks=self.ticks[:n_ticks] + (1. if add else -1.)*other.ticks[:n_ticks] * scale,
                            tick_times=self.tick_times[:n_ticks],
                            name=self.name,
                            live_times=self.live_times[:n_ticks],
                            real_times=self.real_times[:n_ticks],
                            msg_times=None if self.msg_times is None else self.msg_times[:n_ticks],
                            blank_rate=None if self.blank_rate is None else self.blank_rate[:n_ticks],
                            errors=np.sqrt(self.errors[:n_ticks] ** 2 + other.errors[:n_ticks] ** 2 * scale),
                            start_time=self.start_time)

    def simultaneous_addition(self, other: Union[Counts, TickedCounts]):
        """
        Method which simultaneously adds a measurement to the TickedCounts.

        Parameters
        ----------
        other : Counts or TickedCounts
            Signal to be added.

        Returns
        -------
        TickedCounts
        """
        if isinstance(other, Counts):
            return self._simultaneous_op_counts(other, add=True)
        else:
            return self._simultaneous_op_ticked_counts(other, add=True)

    def simultaneous_subtraction(self, other: Union[Counts, TickedCounts]):
        """
        Method which simultaneously subtracts a signal from the TickedCounts. E.g. background subtraction.

        Parameters
        ----------
        other : Counts
            Signal to be subtracted.

        Returns
        -------
        TickedCounts
        """
        if isinstance(other, Counts):
            return self._simultaneous_op_counts(other, add=False)
        else:
            return self._simultaneous_op_ticked_counts(other, add=False)

    def consecutive_addition(self, other: TickedCounts) -> TickedCounts:
        """
        Method which consecutively sums two TickedCounts instances into a new TickedCounts instance.

        Parameters
        ----------
        other : TickedCounts
            TickedCounts instance to sum onto the end of this TickedCounts instance.

        Returns
        -------
        TickedCounts
        """
        if isinstance(other, Counts):
            raise TypeError('Consecutive addition between TickedCounts and Counts instances makes no real sense...')
        return TickedCounts(ticks=np.concatenate((self.ticks, other.ticks)),
                            tick_times=np.concatenate((self.tick_times, other.tick_times)),
                            live_times=np.concatenate((self.live_times, other.live_times)),
                            real_times=np.concatenate((self.real_times, other.real_times)),
                            name=self.name,
                            msg_times=None if self.msg_times is None or other.msg_times is None else np.concatenate(
                                (self.msg_times, other.msg_times)
                            ),
                            blank_rate=None if self.blank_rate is None or other.blank_rate is None else np.concatenate(
                                (self.blank_rate, other.blank_rate)
                            ),
                            errors=None if self._errors is None and other._errors is None else np.concatenate(
                                (self.errors, other.errors)
                            ),
                            start_time=self.start_time,
                            comment=self.comment)


class TickedCountsArray(TickedArrayBase):
    """
    Array of TickedCounts instances.

    Attributes
    ----------
    elements : OrderedDict[str, Measurement]
        OrderedDict, mapping measurements by their names.

    Methods
    -------
    tick_times() -> Array[datetime]
        Property which returns a list of datetime instances labelling the time at which each tick occurred.
    n_ticks() -> int
        Property which returns the number of ticks in the first member of the array.
    n_ticks_all() -> List[int]
        Property which, like n_ticks, returns the number of ticks, but returns a list of values for each member, in
        case you suspect they may not all contain equal numbers.
    names() -> List[str]
        Property which returns a list of the measurement names which are keys to the mapping.
    number_of_members() -> int
        Property which returns the number of members of the array.
    start_times() -> List[datetime]
        Property which returns a list of datetime instances which give the times at which each member of the array
        began to be collected.
    contains_all(names: List[str]) -> bool
        Method which returns True if all the strings in the argument match the names of elements in the array.
    names_match(other: ArrayBase) -> bool
        Method which returns True if all the names of the other array are present in this array, and vice-versa.
    consecutive_addition(other: TickedCountsArray) -> TickedCountsArray
        Method which returns the consecutive sum of the CountsArray with another Array.
    simultaneous_addition(other: TickedCountsArray) -> TickedCountsArray
        Method which returns the simultaneous sum of the CountsArray with another Array.
    consecutive_subtraction(other: CountsArray) -> TickedCountsArray
        Method which returns the consecutive subtraction of another CountsArray from the Array.
    simultaneous_subtraction(other: CountsArray) -> TickedCountsArray:
        Method which returns the simultaneous subtraction of another CountsArray from the Array.
    subtract_backgrounds(self, backgrounds: CountsArray) -> TickedCountsArray:
        Method which subtracts the background from each member of the array, using the array of background counts
        supplied as the argument.
    simultaneous_sum() -> TickedCounts
        Simultaneously sum all members of the array.
    consecutive_sum() -> CountsArray
        Sum all ticks into a CountsArray.
    resample(factor: float) -> TickedCountsArray
        Method which returns a new array with each member resampled after scaling by the factor parameter.
    masked_array(names: List[str]) -> TickedCountsArray
        Method which returns a sub-array of this TickedCountsArray. The method returns a new TickedCountsArray
        containing only TickedCounts instances with the names given.
    get_single_tick(tick_num: int) -> CountsArray
        Method which returns a single tick of Counts from each TickedCounts instance in the array, returning a
        CountsArray.
    sum_ticks(start: datetime, end: datetime) -> CountsArray
        Method which sums the ticks in the TickedCountsArray into a CountsArray.
    simple_windowed_counts_arrays(tick_window_width: int = 1, mode: int = 0) -> Iterator[CountsArray]
        Generator for rolling fixed-length windows. Generator yields CountsArrays.
    plot()
        Method which plots each TickedCounts as a function of time, returning the figure and axes.
    """
    def __init__(self, t_counts: List[TickedCounts]):
        super().__init__(t_counts)

    def __getitem__(self, item: str) -> TickedCounts:
        return super().__getitem__(item)

    def get_single_tick(self, tick_num: int) -> CountsArray:
        """
        Method which returns a single tick of Counts from each TickedCounts instance in the array, returning a
        CountsArray.

        Parameters
        ----------
        tick_num : int
            Index of the tick to be returned.

        Returns
        -------
        CountsArray
        """
        return CountsArray([self[det].get_single_tick(tick_num) for det in self.names])

    def sum_ticks(self, start: Union[datetime, None] = None, end: Union[datetime, None] = None) -> CountsArray:
        """
        Method which sums the ticks in the TickedCountsArray into a CountsArray.

        Parameters
        ----------
        start : datetime
        end : datetime

        Returns
        -------
        CountsArray
        """
        return CountsArray([tc.sum_ticks(start, end) for tc in self])

    def simple_windowed_counts_arrays(self, tick_window_width: int = 1, mode: int = 0) -> Iterator[CountsArray]:
        """
        Generator for rolling fixed-length windows. Generator yields CountsArrays.

        Parameters
        ----------
        tick_window_width : int
            Length of the rolling window, in numbers of ticks.
        mode : int
            Mode (0, 1, or 2)  of the rolling windows.
                0 - windows are always full. For a TickedSpectrum with N ticks, there will be N - tick_window_width + 1
                    spectra yielded.
                    |====================================================| Ticks
                    |====|----------------------------------------->|====| Windows
                1 - windows are always at least half full. For a TickedSpectrum with N ticks, there will be N spectra
                    yielded.
                      |====================================================| Ticks
                    |====|--------------------------------------------->|====| Windows
                2 - windows are always at least one tick full. For a TickedSpectrum with N ticks, there will be
                    N + tick_window_width - 1 spectra yielded.
                         |====================================================| Ticks
                    |====|--------------------------------------------------->|====| Windows

        Yields
        ------
        CountsArray
        """
        generators = [self[d].simple_windowed_counts(tick_window_width, mode) for d in self.names]
        while 1:
            try:
                spectra = [next(gen) for gen in generators]
            except StopIteration:
                return
            yield CountsArray(spectra)

    def masked_array(self, names: List[str]) -> TickedCountsArray:
        """
        Method which returns a sub-array of this TickedCountsArray. The method returns a new TickedCountsArray
        containing only TickedCounts instances with the names given.

        Parameters
        ----------
        names : List[str]
            List of detector names.

        Returns
        -------
        TickedCountsArray
        """
        return self.__class__([self[name] for name in names])

    def plot(self, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """
        Method which plots each TickedCounts as a function of time, returning the figure and axes.

        Parameters
        ----------
        xmin : Union[float, int, None]
            Lower bound of the X axis.
        xmax : Union[float, int, None]
            Upper bound of the X axis.
        ymin : Union[float, int, None]
            Lower bound of the Y axis.
        ymax : Union[float, int, None]
            Upper bound of the Y axis.
        axes : Union[plt.Axes, None]
            Optional axes on which to plot the data. If None, the method generates a new figure and axes.
        figsize : Tuple[float, float]
            Tuple containing the dimensions of the figure if the figure is generated in the method.
        xlabel : str
            Text label for the X axis.
        ylabel : str
            Text label for the Y axis.
        title : Union[str, None]
            Optional title for the axes.
        dpi : int
            Dots per inch value for the figure, if generated by the method.
        logy : bool
            False by default. Set to true to scale the Y-axis logarithmically.
        show : bool
            True by default. Set to False to not automatically show the figure.

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
        """
        return TickedCountsPlotter(ticked_counts=self.elements.values(), **kwargs).plot()

    def subtract_backgrounds(self, bg_array: CountsArray) -> TickedCountsArray:
        """
        Method which subtracts a background CountsArray.

        Parameters
        ----------
        bg_array : CountsArray

        Returns
        -------
        TickedCountsArray
        """
        return self.simultaneous_subtraction(bg_array)

    def simultaneous_addition(self, other: Union[CountsArray, TickedCountsArray]) -> TickedCountsArray:
        """
        Method which simultaneously adds a CountsArray or TickedCountsArray to this TickedCountsArray.
        Not sure why you'd want to though...

        Parameters
        ----------
        other : CountsArray

        Returns
        -------
        TickedCountsArray
        """
        assert self.names_match(other)
        return TickedCountsArray([self[det].simultaneous_addition(other[det]) for det in self.names])

    def simultaneous_subtraction(self, other: Union[CountsArray, TickedCountsArray]) -> TickedCountsArray:
        """
        Method which simultaneously subtracts a CountsArray from this TickedCountsArray. E.g. Background subtraction.

        Parameters
        ----------
        other : CountsArray

        Returns
        -------
        TickedCountsArray
        """
        assert self.names_match(other)
        return TickedCountsArray([self[det].simultaneous_subtraction(other[det]) for det in self.names])

    def consecutive_addition(self, other: TickedCountsArray) -> TickedCountsArray:
        """
        Method which consecutively adds a TickedCountsArray to this TickedCountsArray. Not sure why you'd want to
        though...

        Parameters
        ----------
        other : CountsArray

        Returns
        -------
        TickedCountsArray
        """
        return super(TickedCountsArray, self).consecutive_addition(other)

    def consecutive_sum(self) -> TickedCounts:
        """
        Consecutive sum of each member of the array.

        Returns
        -------
        TickedCounts
        """
        result: Union[TickedCounts, None] = None
        for ticked_counts in self:
            if result is None:
                result = ticked_counts
            else:
                result = result.consecutive_addition(ticked_counts)
        return result

    def simultaneous_sum(self) -> TickedCounts:
        """
        Simultaneous sum tick-wise of each member of the array.

        Returns
        -------
        TickedCounts
        """
        result: Union[TickedCounts, None] = None
        for ticked_counts in self:
            if result is None:
                result = ticked_counts
            else:
                result = result.simultaneous_addition(ticked_counts)
        return result

    def resample(self, factor: float) -> TickedCountsArray:
        """
        Method which returns a new array with each member resampled after scaling by the factor parameter.

        Parameters
        ----------
        factor : float
            Scale by which the measurements will be scaled before resampling.

        Returns
        -------
        TickedCountsArray
        """
        return super(TickedCountsArray, self).resample(factor)



from __future__ import annotations
import numpy as np
import logging
import os
from glob import glob
from typing import NoReturn, Union, List, Tuple, Type, Dict
from scipy.stats import norm as normal_dist
from copy import deepcopy

from geiger.core.calibration import Calibration
from geiger.core.spectrum import Spectrum, EmissionSpectrum
from geiger.algorithms.deconvolution import DeconvolutionAlgorithm, MLEM


DeconvolutionAlgo = Type[DeconvolutionAlgorithm]


class SpectralResponseMatrix:
    def __init__(
            self,
            measurement_calibration: Calibration = Calibration(),
            incident_calibration: Calibration = Calibration(),
            q_matrix: np.ndarray = None,
            engbins: Union[np.ndarray, None] = None,
            efficiency_data: Union[np.ndarray, None] = None,
            photons_per_engbin: float = 1.,
            additional_data: Union[Dict, None] = None
    ):
        self.m_calibration = measurement_calibration
        self.i_calibration = incident_calibration
        if q_matrix is None:
            self.q_matrix = np.zeros((self.i_calibration.n_channels, self.m_calibration.n_channels))
        else:
            self.q_matrix = q_matrix

        if engbins is None:
            self.engbins = np.zeros((self.i_calibration.n_channels, self.m_calibration.n_channels))
        else:
            self.engbins = engbins

        if efficiency_data is None:
            self.eff_data = np.zeros(self.i_calibration.n_channels)
        else:
            self.eff_data = efficiency_data
        self.photons_per_engbin = photons_per_engbin
        self.lld = 0

        self.additional_info = additional_data

        self.log = logging.getLogger(self.__class__.__name__)

    def set_engbin(self, engbin: np.ndarray, engbin_num: int) -> NoReturn:
        if engbin.size != self.m_calibration.n_channels:
            self.log.error(f"Engbin the wrong size. Got an array of size {engbin.size}. "
                           f"Expected {self.m_calibration.n_channels}")
            raise Exception("Engbin wrong size!")
        self.engbins[engbin_num, :] = engbin[:]

    def set_q_response(self, response: np.ndarray, response_num: int) -> NoReturn:
        if response.size != self.m_calibration.n_channels or len(response.shape) != 1:
            self.log.error(f"Response the wrong size. Got an array of shape {response.shape}. "
                           f"Expected {self.m_calibration.n_channels}")
            raise Exception("Response wrong size!")
        self.engbins[response_num, :] = response[:]

    def engbin_as_spectrum(self, engbin_num: int) -> Spectrum:
        return Spectrum(self.m_calibration, self.engbins[engbin_num])

    def q_response_as_spectrum(self, q_response_num: int) -> Spectrum:
        return Spectrum(self.m_calibration, self.q_matrix[q_response_num])

    def __broaden(self, sigmas: np.ndarray) -> NoReturn:
        """ Broaden engbins with sigma (gaussian standard deviation) values
        in units of n_channels numbers for each channel. Saves broadened
        responses to self.qmatrix, overwriting previous values.

        Parameters
        ----------
        sigmas : numpy.ndarray
            Array of sigma values.
        """
        if (self.engbins == 0.).all():
            self.log.warning("You are trying to broaden an empty engbin matrix...")

        def norm(mu: float, sigma: float, num: int):
            x = np.arange(num+1)
            cdf = normal_dist.cdf(x - 0.5, loc=mu, scale=sigma)
            return cdf[1:] - cdf[:-1]

        self.q_matrix[:, :] = 0.
        for ch, sig in enumerate(sigmas):
            broad = norm(ch, sig, self.m_calibration.n_channels)
            for i in range(self.i_calibration.n_channels):
                self.q_matrix[i] += self.engbins[i, ch] * broad

    def broaden(self, fwhms: np.ndarray) -> NoReturn:
        """ Broaden engbins with fwhm values in units of energy for each
        channel. Saves broadened responses to self.qmatrix, overwriting
        previous values.

        Parameters
        ----------
        fwhms : numpy.ndarray
            Array of fwhm values.
        """
        de = self.m_calibration.bin_widths
        sigmas = np.absolute(fwhms / (2.35 * de))
        self.__broaden(sigmas)

    def broaden_poly(self, fwhm_poly: Union[np.ndarray, List]) -> NoReturn:
        """ Broaden engbins with gaussians whose FWHM vary as a quadratic
        polynomial as a function of energy. Saves broadened responses to
        self.qmatrix, overwriting previous values. The FWHM as a function of
        energy is specified by a polynomial equation with coefficients in
        fwhm_poly.

        Parameters
        ----------
        fwhm_poly : list or numpy.ndarray
            quadratic coefficients (c0, c1, c2, c3),

                where FWHM(E) = c0 + c1*E + c2*E**2 + c3*E**3
        """
        energies = self.m_calibration.energies[:, np.newaxis]
        powers = np.arange(4)[np.newaxis, :]
        fwhm_poly = np.array(fwhm_poly)[np.newaxis, :]
        fwhms = (fwhm_poly[3] * np.power(energies, powers)).sum(axis=1)

        self.broaden(fwhms)

    def broaden_exp(self, fwhm_exp: Union[np.ndarray, List]) -> NoReturn:
        """ Broaden engbins with gaussians whose FWHM vary as a exponential as
        a function of energy. Saves broadened responses to self.qmatrix,
        overwriting previous values. The FWHM as a function of energy is
        specified by an exponential equation with coefficients in
        fwhm_exp.

        Parameters
        ----------
        fwhm_exp : list or numpy.ndarray
            exponential coefficients (c0, c1, c2),

                where FWHM(E) = c0 + c1*E**c2
        """

        fwhms = fwhm_exp[1] * self.m_calibration.energies ** fwhm_exp[2]
        fwhms[np.isnan(fwhms)] = 1E-20
        fwhms += fwhm_exp[0]

        self.broaden(fwhms)

    def set_photons_per_engbin(self, photons: float) -> NoReturn:
        """ Function that sets the number of
        photons incident on the detector per engbin. This number is used to
        normalise the engbins before convolutions/deconvolutions.

        Parameters
        ----------
        photons : int
            Number of photons per engbin.
        """
        self.photons_per_engbin = photons

    def set_lld(self, lld_energy: float) -> NoReturn:
        """ Function which sets the lower level discriminator, in units of
        energy.

        Parameters
        ----------
        lld_energy : float
            Low level discriminator in units of energy.
        """
        for i in range(self.m_calibration.n_channels):
            if lld_energy < self.m_calibration.energies[i]:
                self.lld = i
                return

    def get_lld(self) -> float:
        """ Function which returns the low level discriminator for the
        convolution and deconvolution algorithms, in units of energy.

        Returns
        -------
        float
            Lld in units of energy.
        """
        return self.m_calibration.energies[self.lld]

    def generate_eff_data(self) -> NoReturn:
        """ Function which generates efficiency data for saving to sqf
        """
        self.eff_data = self.engbins.sum(axis=1) / self.photons_per_engbin

    def convolve_spectrum(self, spectrum: Spectrum, matrix: str = "q_matrix", with_errors: bool = False) -> Spectrum:
        """ Function which convolves an incident spectrum with the Q-matrix,
        and returns the convolved spectrum.

        Parameters
        ----------
        spectrum : Spectrum
            Spectrum to be convolved. Note, spectrum will be rebinned if not
            same calibration and size as the array of responses in the
            Q-matrix.
        matrix : str
            Either 'q_matrix' (default) or 'engbins'
        with_errors : bool
            Convolve the uncertainties in the input spectrum into the convolved spectrum.

        Returns
        -------
        Instance of Spectrum class
            Convolution of incident spectrum with response-matrix.
        """
        
        # Check if same number of n_channels and calibration.
        if not spectrum.calibration == self.i_calibration:
            spectrum = spectrum.rebin(self.i_calibration)

        mat = getattr(self, matrix)
        normalised_responses = (mat.T / self.photons_per_engbin).copy()
        normalised_responses[:int(self.lld)] = 0
        # Generate convolution spectrum
        name = spectrum.name + " " if spectrum.name is not None else ""
        ret_spec = Spectrum(self.m_calibration.energies,
                            np.dot(normalised_responses, spectrum.counts),
                            live_time=spectrum.live_time,
                            real_time=spectrum.real_time,
                            error=np.sqrt(np.dot(normalised_responses**2, spectrum.error**2)) if with_errors else None,
                            name=name + "convolved",
                            comment=spectrum.comment,
                            start_time=spectrum.start_time)
        return ret_spec

    def generate_spectrum_from_lara_text(self,
                                         lara_txt_file: str,
                                         n: int = 1.0E6) -> Spectrum:
        """ Method for generating a modelled spectrum from lara isotope data.
        Data can be downloaded here:
            http://www.nucleide.org/Laraweb/index.php

        Parameters
        ----------
        lara_txt_file : str
            Path to the lara text file.
        n : int
            Number of photons incident on the detector

        Returns
        -------
            Spectrum
        """
        inc = EmissionSpectrum.from_lara_decay_data_file(lara_txt_file, self.i_calibration)
        inc.counts *= n
        return self.convolve_spectrum(inc)

    def generate_spectrum_from_lara_online(self,
                                           isotope: str = "Cs-137",
                                           n: int = 1.0E6) -> Spectrum:
        """ Method for generating a modelled spectrum from lara isotope data.
        Data can be downloaded here:
            http://www.nucleide.org/Laraweb/index.php

        Parameters
        ----------
        isotope : str
            Isotope in this format: "Cs-137"
        n : int
            Number of photons incident on the detector

        Returns
        -------
            Spectrum
        """
        inc = EmissionSpectrum.from_lara_online(isotope, self.i_calibration)
        inc.counts *= n
        return self.convolve_spectrum(inc)

    def rebin_measurement_axis(self, new_calibration):
        new_qmatrix = deepcopy(self)
        new_qmatrix.q_matrix = np.zeros((self.i_calibration.n_channels, new_calibration.n_channels))
        new_qmatrix.engbins = np.zeros((self.i_calibration.n_channels, new_calibration.n_channels))
        for i in range(new_qmatrix.i_calibration.n_channels):
            engbin = self.engbin_as_spectrum(i)
            new_qmatrix.engbins[i, :] = engbin.rebin(new_calibration).counts[:]
            q_response = self.q_response_as_spectrum(i)
            new_qmatrix.q_matrix[i, :] = q_response.rebin(new_calibration).counts[:]
        new_qmatrix.m_calibration = new_calibration
        return new_qmatrix

    def find_deconvolution_channel(self, energy: float) -> int:
        return self.i_calibration.find_channel(energy)

    def find_measurement_channel(self, energy: float) -> int:
        return self.m_calibration.find_channel(energy)

    def deconvolve_spectrum(
            self,
            spectrum: Spectrum,
            iterations: int = 1000,
            init: np.ndarray = None,
            method: DeconvolutionAlgo = MLEM
    ) -> Tuple[Spectrum, Spectrum]:
        """ Method for deconvolving a spectrum using this Q-matrix. Uses the Shepp and Vardi MLEM algorithm by default.

        Parameters
        ----------
        spectrum : Spectrum
            Spectrum to be deconvolved
        iterations: int
        init : np.ndarray
        method: DeconvolutionAlgorithm

        Returns
        -------
        Tuple[Spectrum, Spectrum]
        """
        # Check if same number of n_channels and calibration.
        if not spectrum.calibration == self.m_calibration:
            spectrum = spectrum.rebin(self.m_calibration)

        normalised_responses = self.q_matrix.T / (self.photons_per_engbin if self.photons_per_engbin > 0 else 1.)
        normalised_responses[:int(self.lld)] = 0
        name = spectrum.name + " " if spectrum.name is not None else ""

        deconvolver = method(normalised_responses)

        decon_counts, recon_counts = deconvolver.deconvolve_measurement(spectrum.counts, iterations, init)

        decon_spec = Spectrum(self.i_calibration.energies,
                              decon_counts,
                              name=name + "deconvolved",
                              live_time=spectrum.live_time,
                              real_time=spectrum.real_time)
        recon_spec = Spectrum(self.m_calibration.energies,
                              recon_counts,
                              name=name + "reconvolved",
                              live_time=spectrum.live_time,
                              real_time=spectrum.real_time)
        return decon_spec, recon_spec

    def bootstrap_deconvolve_spectrum(
            self,
            spectrum: Spectrum,
            iterations: int = 1000,
            bootstraps: int = 100,
            init: np.ndarray = None,
            method: DeconvolutionAlgo = MLEM,
            progress: bool = False
    ) -> Tuple[Spectrum, Spectrum]:
        # Check if same number of n_channels and calibration.
        if not spectrum.calibration == self.m_calibration:
            spectrum = spectrum.rebin(self.m_calibration)

        normalised_responses = self.q_matrix.T / (self.photons_per_engbin if self.photons_per_engbin > 0 else 1.)
        normalised_responses[:int(self.lld)] = 0
        name = spectrum.name + " " if spectrum.name is not None else ""

        deconvolver = method(normalised_responses)

        decon_counts, recon_counts, decon_err, recon_err = deconvolver.bootstrap_deconvolve_measurement(
            spectrum.counts,
            spectrum.error,
            iterations,
            bootstraps,
            init,
            progress
        )

        decon_spec = Spectrum(self.i_calibration.energies,
                              decon_counts,
                              name=name + "deconvolved",
                              live_time=spectrum.live_time,
                              real_time=spectrum.real_time,
                              error=decon_err)
        recon_spec = Spectrum(self.m_calibration.energies,
                              recon_counts,
                              name=name + "reconvolved",
                              live_time=spectrum.live_time,
                              real_time=spectrum.real_time,
                              error=recon_err)
        return decon_spec, recon_spec

    @classmethod
    def from_engbin_directory(cls, engbin_directory: str, incident_calibration: Calibration,
                              measurement_calibration: Calibration = None, energy_column: int = 0,
                              counts_column: int = 1, start_index: int = 0):
        engbin_files = glob(os.path.join(engbin_directory, "engbin*.eb"))
        q_matrix = None
        for i in range(start_index, start_index+len(engbin_files)):
            engbin_spectrum = Spectrum.from_txt(os.path.join(engbin_directory, f"engbin{i}.eb"),
                                                energy_column=energy_column, counts_column=counts_column)
            if q_matrix is None:
                q_matrix = cls(engbin_spectrum.calibration if measurement_calibration is None
                               else measurement_calibration, incident_calibration)
            q_matrix.set_engbin(engbin_spectrum.counts, i - start_index)
        if q_matrix is None:
            raise ValueError(f"No engbins found in {engbin_directory}")
        return q_matrix



__version__ = '0.0.3'

from .core.calibration import Calibration
from .core.counts import Counts, CountsArray, TickedCounts, TickedCountsArray
from .core.spectrum import Spectrum, SpectrumArray, TickedSpectrum, TickedSpectrumArray

from .algorithms.spectral_response_matrix import SpectralResponseMatrix
from .algorithms.deconvolution import MLEM



from __future__ import annotations
import numpy as np
from typing import Union, List, Tuple


class Calibration:
    """
    A class containing calibration information for spectra or Q-matrices.

    This class handles all matters relating to calibrations, energy binning etc.

    Attributes
    ----------
    n_channels : int
        Number of channels
    coefficients : Array[length = 3]
        Calibration coefficients
    label : str
        Label for the energy scale, used for plotting. Default is "Energy, keV".

    Methods
    -------
    is_calibrated -> bool
        Property which returns True if calibration coefficients provided.
    channel_numbers -> np.ndarray
        Property which returns a 1D array of channel numbers for the calibration.
    energies -> np.ndarray
        Property which returns an array of energy values for the lower limits of each energy bin, if the calibration
        coefficients are supplied, otherwise it returns the channel numbers.
    bin_widths -> np.ndarray or None.
        Property which returns an array of bin widths in energy units if calibration coefficients are supplied,
        otherwise it returns None.
    lower_bin_limits -> np.ndarray or None
        Property which returns an array of energy values for the lower limits of each energy bin, if the calibration
        coefficients are supplied, otherwise it returns None.
    upper_bin_limits -> np.ndarray or None
        Property which returns an array of energy values for the upper limits of each energy bin, if the calibration
        coefficients are supplied, otherwise it returns None.
    all_bin_limits -> np.ndarray or None
        Property which returns an array of energy values for the upper and lower limits of each energy bin, if the
        calibration coefficients are supplied, otherwise it returns None.
    centre_bin_energies -> np.ndarray or None
        Property which returns an array of energy values for the centre of each energy bin, if the calibration
        coefficients are supplied, otherwise it returns None.
    find_channel(energy: float) -> int or None
        Method which returns the channel number in which a deposit of energy 'energy' would be counted. If calibration
        coefficients are not provided, or if the energy is outside the bounds of the energy scale, this method returns
        None.
    energy_mask(lower: float, upper: float) -> np.ndarray or None
        Method which returns a boolean array which can be applied as a boolean mask on arrays of the same length. The
        'lower' and 'upper' parameters set the lower and upper limits on the energy scale for which the mask is True.
        If no calibration coefficients are supplied, this method returns None.
    from_energies(energies: np.ndarray) -> Calibration
        Class method which returns a Calibration instance calibrated on the energy values in the 'energy' array.
    channel_numbers_only(n_channels: int) -> Calibration
        Class method which returns a Calibration instance without calibration coefficients, with label equal to
        'Channel #'
    """
    def __init__(
            self,
            n_channels: int = 1024,
            coefficients: Union[np.ndarray, Tuple[float, float, float], List[float], None] = None,
            label: Union[str, None] = None
    ):
        """
        Parameters
        ----------
        n_channels : int
            Number of channels, default = 1024.
        coefficients : Tuple[float, float, float]
            offset, linear and quadratic calibration coefficients which define the lower energy bound of each channel
        label : Optional[str]
            Label for the energy scale. Used when plotting. Default is "Energy, keV". Use this if energy is in different
            units.
        """

        self.n_channels = n_channels
        self._is_calibrated = coefficients is not None
        if self.is_calibrated:
            self.coefficients: Union[np.ndarray, None] = np.array(coefficients)
        else:
            self.coefficients: Union[np.ndarray, None] = None
        self.label = label if label is not None else 'Energy, keV'

    @property
    def is_calibrated(self) -> bool:
        """
        Property which returns True if calibration coefficients provided.

        Returns
        -------
        bool
        """
        return self._is_calibrated

    @property
    def channel_numbers(self) -> np.ndarray:
        """
        Property which returns a 1D array of channel numbers for the calibration.

        Returns
        -------
        np.ndarray
        """
        return np.arange(self.n_channels)

    @property
    def bin_widths(self) -> Union[np.ndarray, None]:
        """
        Property which returns an array of bin widths in energy units if calibration coefficients are supplied,
        otherwise it returns None.

        Returns
        -------
        np.ndarray or None
        """
        if self.is_calibrated:
            return self.upper_bin_limits - self.lower_bin_limits
        else:
            return None

    @property
    def lower_bin_limits(self) -> Union[np.ndarray, None]:
        """
        Property which returns an array of energy values for the lower limits of each energy bin, if the calibration
        coefficients are supplied, otherwise it returns None.

        Returns
        -------
        np.ndarray or None
        """
        if self.is_calibrated:
            return self.all_bin_limits[:-1]
        else:
            return None

    @property
    def upper_bin_limits(self) -> Union[np.ndarray, None]:
        """
        Property which returns an array of energy values for the upper limits of each energy bin, if the calibration
        coefficients are supplied, otherwise it returns None.

        Returns
        -------
        np.ndarray or None
        """
        if self.is_calibrated:
            return self.all_bin_limits[1:]
        else:
            return None

    @property
    def all_bin_limits(self) -> Union[np.ndarray, None]:
        """
        Property which returns an array of energy values for the upper and lower limits of each energy bin, if the
        calibration coefficients are supplied, otherwise it returns None.

        Returns
        -------
        np.ndarray or None
        """
        if self.is_calibrated:
            channels = np.arange(0., self.n_channels + 1)
            powers = np.power(channels[:, np.newaxis], np.arange(3)[np.newaxis, :])
            return (powers * self.coefficients[np.newaxis, :]).sum(axis=1)
        else:
            return None

    @property
    def centre_bin_energies(self) -> Union[np.ndarray, None]:
        """
        Property which returns an array of energy values for the centre of each energy bin, if the calibration
        coefficients are supplied, otherwise it returns None.

        Returns
        -------
        np.ndarray or None
        """
        if self.is_calibrated:
            return (self.lower_bin_limits + self.upper_bin_limits) / 2.
        else:
            return None

    @property
    def energies(self) -> np.ndarray:
        """
        Property which returns an array of energy values for the lower limits of each energy bin, if the calibration
        coefficients are supplied, otherwise it returns the channel numbers.

        Returns
        -------
        np.ndarray
        """
        if self.is_calibrated:
            return self.lower_bin_limits
        else:
            return self.channel_numbers

    def find_channel(self, energy: float) -> Union[int, None]:
        """
        Method which returns the channel number in which a deposit of energy 'energy' would be counted. If calibration
        coefficients are not provided, or if the energy is outside the bounds of the energy scale, this method returns
        None.

        Parameters
        ----------
        energy : float
            Energy in the same units as the calibration coefficients.

        Returns
        -------
        int or None
        """
        if self.is_calibrated:
            arr = np.arange(self.n_channels)[np.logical_and(self.lower_bin_limits <= energy,
                                                            self.upper_bin_limits > energy)]
            if arr.size == 0:
                return None
            else:
                return int(arr[0])
        else:
            return None

    def energy_mask(self, lower: float = None, upper: float = None) -> Union[np.ndarray, None]:
        """
        Method which returns a boolean array which can be applied as a boolean mask on arrays of the same length. The
        'lower' and 'upper' parameters set the lower and upper limits on the energy scale for which the mask is True.
        If no calibration coefficients are supplied, this method returns None.

        Parameters
        ----------
        lower : float or None
            Lower energy limit for the boolean mask
        upper : float or None
            Upper energy limit for the boolean mask

        Returns
        -------
        np.ndarray
        """
        if self.is_calibrated:
            mask = np.ones(self.n_channels, dtype=bool)
            if upper is not None:
                mask = np.logical_and(mask, self.energies < upper)
            if lower is not None:
                mask = np.logical_and(mask, self.energies >= lower)
            return mask
        else:
            return None

    def __eq__(self, other: Calibration) -> bool:
        return np.isclose(self.coefficients, other.coefficients).all() and self.n_channels == other.n_channels

    @classmethod
    def from_energies(cls, energies: np.ndarray, bin_label: str = 'lower') -> Calibration:
        """
        Class method which returns a Calibration instance calibrated on the energy values in the 'energy' array.

        Parameters
        ----------
        energies : np.ndarray
            Array of energy values from which to create a calibration object.
        bin_label : str
            Either 'upper', 'middle' or 'lower', specifying where in each bin the energy array is labelling.
        """

        upper = bin_label == 'upper'
        middle = bin_label == 'middle'
        lower = bin_label == 'lower'

        assert isinstance(energies, np.ndarray)
        try:
            assert upper or middle or lower
        except AssertionError as _:
            raise AssertionError(f"'bin_label' keyword argument to Classes.Calibration must be either 'upper', "
                                 f"'middle', or 'lower'. You gave '{bin_label}'")

        channels = energies.size
        grad = np.gradient(energies)
        linear = np.array_equal(grad[1:], grad[:-1])

        if middle:
            energies -= grad/2.
        elif upper:
            energies -= grad

        if linear:
            coefficients = np.array([energies[0], grad[0], 0])
        else:
            coefficients = np.polyfit(np.arange(channels), energies, 2)[::-1]

        return cls(channels, coefficients)

    @classmethod
    def channel_numbers_only(cls, n_channels: int) -> Calibration:
        """
        Class method which returns a Calibration instance without calibration coefficients, with label equal to
        'Channel #'

        Parameters
        ----------
        n_channels : int
            Number of channels.

        Returns
        -------
        Calibration
        """
        return cls(n_channels=n_channels, label='Channel #')



# -*- coding: utf-8 -*-
"""
Created on Tue Aug  7 18:24:00 2018

@author: Ian Della-Rocca
"""

import matplotlib.pyplot as plt
import matplotlib.dates as mpl_dates
from dataclasses import dataclass, field
from typing import Union, NewType, Tuple, List, Any
from datetime import datetime, timedelta
from abc import ABC

Number = NewType('Number', Union[int, float, None])
Spectrum = NewType('Spectrum', Any)
TickedCounts = NewType('TickedCounts', Any)
HealthLibrary = NewType('HealthLibrary', Any)
HealthTracker = NewType('HealthTracker', Any)


@dataclass
class PlotterBase:
    xmin: Number = 0.
    xmax: Number = None
    ymin: Number = 0.
    ymax: Number = None
    xlabel: str = 'Energy [keV]'
    ylabel: str = 'Counts'
    title: str = None
    figsize: Tuple[float, float] = (9., 5.)
    dpi: int = 80
    logy: bool = False
    v_lines: List[float] = None
    h_lines: List[float] = None
    v_line_col: str = 'r'
    h_line_col: str = 'b'
    show: bool = True
    tight: bool = False
    save: str = None
    close: bool = False
    axis: Union[plt.Axes, None] = None
    legend: bool = True
    colors: List[str] = None

    figure: plt.Figure = field(init=False,  default=None)

    def __post_init__(self):
        if self.colors is None:
            self.colors = [f'C{i}' for i in range(10)]
        if self.axis is None:
            self.figure, self.axis = plt.subplots(figsize=self.figsize, dpi=self.dpi)
        else:
            self.figure = self.axis.figure

        self.axis.set_xlabel(self.xlabel)
        self.axis.set_ylabel(self.ylabel)

        self.axis.set_title(self.title)

    def plot(self):
        self._do_plotting()
        self._finalize_plot()
        return self.figure, self.axis

    def _do_plotting(self):
        raise NotImplementedError

    def _finalize_plot(self):
        if self.legend:
            self.axis.legend()

        self.axis.set_xlim(left=self.xmin, right=self.xmax)
        self.axis.set_ylim(bottom=self.ymin, top=self.ymax)

        if self.logy:
            self.axis.set_yscale('logy')
            if self.xmin == 0.:
                self.axis.set_ylim(bottom=None)

        if self.v_lines is not None:
            for line in self.v_lines:
                self.axis.axvline(line, color=self.v_line_col)
        if self.h_lines is not None:
            for line in self.h_lines:
                self.axis.axhline(line, color=self.h_line_col)

        if self.tight:
            plt.tight_layout()

        if self.save is not None:
            self.figure.savefig(self.save)

        if self.show:
            plt.show()

        if self.close:
            plt.close(self.figure)


@dataclass
class SpectrumPlotter(PlotterBase):
    spectra: List[Spectrum] = field(default_factory=[])

    def _do_plotting(self):
        for i, spectrum in enumerate(self.spectra):
            self.axis.step(spectrum.energies, spectrum.counts, where='mid', color=self.colors[i], label=spectrum.name)


@dataclass
class TimeSeriesPlotter(PlotterBase, ABC):
    xmin: Union[datetime, None] = None
    xmax: Union[datetime, None] = None
    xlabel: str = 'Time'
    ylabel: str = 'Value'
    figsize: Tuple[float, float] = (14., 7.)
    x_tick_interval: Union[float, None] = None

    earliest: datetime = field(init=False)
    latest: datetime = field(init=False)

    def __post_init__(self):
        super(TimeSeriesPlotter, self).__post_init__()
        self.set_earliest_and_latest()
        if self.x_tick_interval is None:
            threshold1 = 20.
            threshold2 = 100.
            threshold3 = 300.
            threshold4 = 1800.
            threshold5 = 3600.
            threshold6 = 21600.
            threshold7 = 43200.
            threshold8 = 86400.
            total_interval = (self.latest - self.earliest).total_seconds()
            if total_interval <= threshold1:
                tick_interval = 1.0
            elif threshold1 < total_interval <= threshold2:
                tick_interval = 5.0
            elif threshold2 < total_interval <= threshold3:
                tick_interval = 10.0
            elif threshold3 < total_interval <= threshold4:
                tick_interval = 60.0
            elif threshold4 < total_interval <= threshold5:
                tick_interval = 300.0
            elif threshold5 < total_interval <= threshold6:
                tick_interval = 1800.0
            elif threshold6 < total_interval <= threshold7:
                tick_interval = 3600.0
            elif threshold7 < total_interval <= threshold8:
                tick_interval = 7200.0
            else:
                tick_interval = 21600.0
            self.x_tick_interval = tick_interval
        day_of_earliest = datetime(year=self.earliest.year, month=self.earliest.month,
                                   day=self.earliest.day, tzinfo=self.earliest.tzinfo)
        minor_ticks = [self.earliest - timedelta(seconds=((self.earliest - day_of_earliest).total_seconds() %
                                                 self.x_tick_interval - self.x_tick_interval))]
        while minor_ticks[-1] < self.latest:
            minor_ticks.append(minor_ticks[-1] + timedelta(seconds=self.x_tick_interval))
        minor_ticks.pop(-1)
        self.axis.set_xticks([self.earliest, self.latest])
        self.axis.set_xticks(minor_ticks, minor=True)
        # plt.xticks(rotation=90)
        # plt.setp(self.axis.xaxis.get_minorticklabels(), rotation=0)
        plt.setp(self.axis.xaxis.get_majorticklabels(), fontweight='bold')
        if self.x_tick_interval < 60.:
            self.axis.xaxis.set_minor_formatter(mpl_dates.DateFormatter('%H:%M:%S'))
            self.axis.xaxis.set_major_formatter(mpl_dates.DateFormatter('\n%H:%M:%S'))
        else:
            self.axis.xaxis.set_minor_formatter(mpl_dates.DateFormatter('%H:%M:%S'))
            self.axis.xaxis.set_major_formatter(mpl_dates.DateFormatter('\n%x\n%X'))

    def set_earliest_and_latest(self):
        raise NotImplementedError


@dataclass
class TickedCountsPlotter(TimeSeriesPlotter):
    ticked_counts: List[TickedCounts] = field(default_factory=[])
    ylabel: str = 'Counts'

    def set_earliest_and_latest(self):
        self.earliest = min([tc.tick_times[tc.tick_times != None][0] for tc in self.ticked_counts])
        self.latest = max([tc.tick_times[tc.tick_times != None][-1] for tc in self.ticked_counts])

    def _do_plotting(self):
        for i, item in enumerate(self.ticked_counts):
            self.axis.step(item.tick_times, item.ticks, where='post', color=self.colors[i], label=item.name)

